import mpiunittest as unittest

from mpi4py import MPI


class TestRequest(unittest.TestCase):
    #
    def setUp(self):
        self.REQUEST = MPI.Request()
        self.STATUS = MPI.Status()

    def testWait(self):
        self.REQUEST.Wait()
        self.REQUEST.Wait(None)
        self.REQUEST.Wait(self.STATUS)
        self.assertIs(self.REQUEST.Wait(), True)
        self.REQUEST.wait()
        self.REQUEST.wait(None)
        self.REQUEST.wait(self.STATUS)
        self.assertIsNone(self.REQUEST.wait())

    def testTest(self):
        self.REQUEST.Test()
        self.REQUEST.Test(None)
        self.REQUEST.Test(self.STATUS)
        self.assertIs(self.REQUEST.Test(), True)
        self.REQUEST.test()
        self.REQUEST.test(None)
        self.REQUEST.test(self.STATUS)
        self.assertEqual(self.REQUEST.test(), (True, None))

    @unittest.skipMPI("MPICH1")
    def testGetStatus(self):
        try:
            flag = self.REQUEST.Get_status()
        except NotImplementedError:
            self.skipTest("mpi-request-get_status")
        self.assertTrue(flag)
        flag = self.REQUEST.Get_status(self.STATUS)
        self.assertTrue(flag)
        self.assertEqual(self.STATUS.Get_source(), MPI.ANY_SOURCE)
        self.assertEqual(self.STATUS.Get_tag(), MPI.ANY_TAG)
        self.assertEqual(self.STATUS.Get_error(), MPI.SUCCESS)
        self.assertEqual(self.STATUS.Get_count(MPI.BYTE), 0)
        self.assertEqual(self.STATUS.Get_elements(MPI.BYTE), 0)
        try:
            self.assertFalse(self.STATUS.Is_cancelled())
        except NotImplementedError:
            self.skipTest("mpi-status-is_cancelled")
        flag = self.REQUEST.get_status()
        self.assertTrue(flag)
        flag = self.REQUEST.get_status(self.STATUS)
        self.assertEqual(self.STATUS.source, MPI.ANY_SOURCE)
        self.assertEqual(self.STATUS.tag, MPI.ANY_TAG)
        self.assertEqual(self.STATUS.error, MPI.SUCCESS)


class TestRequestArray(unittest.TestCase):
    #
    def setUp(self):
        self.REQUESTS = [MPI.Request() for i in range(5)]
        self.STATUSES = [MPI.Status() for i in range(5)]
        self.STATUS = self.STATUSES[0]
        for status in self.STATUSES:
            status.source = 0
            status.tag = 0
            status.error = MPI.ERR_OTHER

    def testWaitany(self):
        MPI.Request.Waitany(self.REQUESTS)
        MPI.Request.Waitany(self.REQUESTS, None)
        MPI.Request.Waitany(self.REQUESTS, self.STATUSES[0])
        MPI.Request.waitany(self.REQUESTS)
        MPI.Request.waitany(self.REQUESTS, None)
        MPI.Request.waitany(self.REQUESTS, self.STATUSES[0])

    def testTestany(self):
        MPI.Request.Testany(self.REQUESTS)
        MPI.Request.Testany(self.REQUESTS, None)
        MPI.Request.Testany(self.REQUESTS, self.STATUSES[0])
        MPI.Request.testany(self.REQUESTS)
        MPI.Request.testany(self.REQUESTS, None)
        MPI.Request.testany(self.REQUESTS, self.STATUSES[0])

    def testGetStatusAny(self):
        with self.catchNotImplementedError(4, 1):
            status = self.STATUS
            index, flag = MPI.Request.Get_status_any(self.REQUESTS)
            self.assertEqual(index, MPI.UNDEFINED)
            self.assertTrue(flag)
            index, flag = MPI.Request.Get_status_any(self.REQUESTS, None)
            self.assertEqual(index, MPI.UNDEFINED)
            self.assertTrue(flag)
            if unittest.is_mpi("impi(>=2021.14.0)"):
                status.error = MPI.SUCCESS
            index, flag = MPI.Request.Get_status_any(self.REQUESTS, status)
            self.assertEqual(index, MPI.UNDEFINED)
            self.assertTrue(flag)
            self.assertEqual(status.source, MPI.ANY_SOURCE)
            self.assertEqual(status.tag, MPI.ANY_TAG)
            self.assertEqual(status.error, MPI.SUCCESS)
        with self.catchNotImplementedError(4, 1):
            index, flag = MPI.Request.get_status_any(self.REQUESTS)
            self.assertEqual(index, MPI.UNDEFINED)
            self.assertTrue(flag)

    def testWaitall(self):
        MPI.Request.Waitall(self.REQUESTS)
        MPI.Request.Waitall(self.REQUESTS, None)
        self.assertIs(MPI.Request.Waitall(self.REQUESTS), True)
        for statuses in (tuple(self.STATUSES), (self.STATUSES[0],), ()):
            MPI.Request.Waitall(self.REQUESTS, statuses)
        for statuses in (self.STATUSES, []):
            MPI.Request.Waitall(self.REQUESTS, statuses)
            self.assertEqual(len(statuses), len(self.REQUESTS))
        MPI.Request.waitall(self.REQUESTS)
        MPI.Request.waitall(self.REQUESTS, None)
        for statuses in (self.STATUSES, []):
            MPI.Request.waitall(self.REQUESTS, statuses)
            self.assertEqual(len(statuses), len(self.REQUESTS))

    def testTestall(self):
        MPI.Request.Testall(self.REQUESTS)
        MPI.Request.Testall(self.REQUESTS, None)
        self.assertIs(MPI.Request.Testall(self.REQUESTS), True)
        for statuses in (self.STATUSES, []):
            MPI.Request.Testall(self.REQUESTS, statuses)
            self.assertEqual(len(statuses), len(self.REQUESTS))
        MPI.Request.testall(self.REQUESTS)
        MPI.Request.testall(self.REQUESTS, None)
        for statuses in (self.STATUSES, []):
            MPI.Request.testall(self.REQUESTS, statuses)
            self.assertEqual(len(statuses), len(self.REQUESTS))

    def testGetStatusAll(self):
        with self.catchNotImplementedError(4, 1):
            statuses = self.STATUSES
            flag = MPI.Request.Get_status_all(self.REQUESTS)
            self.assertTrue(flag)
            flag = MPI.Request.Get_status_all(self.REQUESTS, None)
            self.assertTrue(flag)
            flag = MPI.Request.Get_status_all(self.REQUESTS, statuses)
            self.assertTrue(flag)
            for status in statuses:
                self.assertEqual(status.source, MPI.ANY_SOURCE)
                self.assertEqual(status.tag, MPI.ANY_TAG)
                self.assertEqual(status.error, MPI.SUCCESS)
        with self.catchNotImplementedError(4, 1):
            flag = MPI.Request.get_status_all(self.REQUESTS)
            self.assertTrue(flag)

    def testWaitsome(self):
        ret = MPI.Request.Waitsome(self.REQUESTS)
        self.assertIsNone(ret)
        ret = MPI.Request.Waitsome(self.REQUESTS, None)
        self.assertIsNone(ret)
        for statuses in (self.STATUSES, []):
            slen = len(statuses)
            ret = MPI.Request.Waitsome(self.REQUESTS, statuses)
            self.assertIsNone(ret)
            self.assertEqual(len(statuses), slen)

    def testTestsome(self):
        ret = MPI.Request.Testsome(self.REQUESTS)
        self.assertIsNone(ret)
        ret = MPI.Request.Testsome(self.REQUESTS, None)
        self.assertIsNone(ret)
        for statuses in (self.STATUSES, []):
            slen = len(statuses)
            ret = MPI.Request.Testsome(self.REQUESTS, statuses)
            self.assertIsNone(ret)
            self.assertEqual(len(statuses), slen)

    def testGetStatusSome(self):
        with self.catchNotImplementedError(4, 1):
            statuses = self.STATUSES
            indices = MPI.Request.Get_status_some(self.REQUESTS)
            self.assertIsNone(indices)
            indices = MPI.Request.Get_status_some(self.REQUESTS, None)
            self.assertIsNone(indices)
            indices = MPI.Request.Get_status_some(self.REQUESTS, statuses)
            self.assertIsNone(indices)
        with self.catchNotImplementedError(4, 1):
            indices = MPI.Request.get_status_some(self.REQUESTS)
            self.assertIsNone(indices)


if __name__ == "__main__":
    unittest.main()



cdef class Message:
    """
    Matched message.
    """

    def __cinit__(self, Message message: Message | None = None):
        cinit(self, message)

    def __dealloc__(self):
        dealloc(self)

    def __richcmp__(self, other, int op):
        if not isinstance(other, Message): return NotImplemented
        return richcmp(self, other, op)

    def __bool__(self) -> bool:
        return nonnull(self)

    def __reduce__(self) -> str | tuple[Any, ...]:
        return reduce_default(self)

    property handle:
        """MPI handle."""
        def __get__(self) -> int:
            return tohandle(self)

    @classmethod
    def fromhandle(cls, handle: int) -> Message:
        """
        Create object from MPI handle.
        """
        return fromhandle(<MPI_Message> <Py_uintptr_t> handle)

    def free(self) -> None:
        """
        Do nothing.
        """
        safefree(self)

    # Matching Probe
    # --------------

    @classmethod
    def Probe(
        cls,
        Comm comm: Comm,
        int source: int = ANY_SOURCE,
        int tag: int = ANY_TAG,
        Status status: Status | None = None,
    ) -> Self:
        """
        Blocking test for a matched message.
        """
        cdef MPI_Message cmessage = MPI_MESSAGE_NULL
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_Mprobe(
            source, tag, comm.ob_mpi, &cmessage, statusp) )
        cdef Message message = <Message>New(cls)
        message.ob_mpi = cmessage
        return message

    @classmethod
    def Iprobe(
        cls,
        Comm comm: Comm,
        int source: int = ANY_SOURCE,
        int tag: int = ANY_TAG,
        Status status: Status | None = None,
    ) -> Self | None:
        """
        Nonblocking test for a matched message.
        """
        cdef int flag = 0
        cdef MPI_Message cmessage = MPI_MESSAGE_NULL
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_Improbe(
             source, tag, comm.ob_mpi, &flag, &cmessage, statusp) )
        if flag == 0: return None
        cdef Message message = <Message>New(cls)
        message.ob_mpi = cmessage
        return message

    # Matched receives
    # ----------------

    def Recv(
        self,
        buf: BufSpec,
        Status status: Status | None = None,
    ) -> None:
        """
        Blocking receive of matched message.
        """
        cdef MPI_Message message = self.ob_mpi
        cdef int source = MPI_ANY_SOURCE
        if message == MPI_MESSAGE_NO_PROC:
            source = MPI_PROC_NULL
        cdef _p_msg_p2p rmsg = message_p2p_recv(buf, source)
        cdef MPI_Status *statusp = arg_Status(status)
        with nogil: CHKERR( MPI_Mrecv_c(
            rmsg.buf, rmsg.count, rmsg.dtype,
            &message, statusp) )
        if self is not __MESSAGE_NO_PROC__:
            self.ob_mpi = message

    def Irecv(self, buf: BufSpec) -> Request:
        """
        Nonblocking receive of matched message.
        """
        cdef MPI_Message message = self.ob_mpi
        cdef int source = MPI_ANY_SOURCE
        if message == MPI_MESSAGE_NO_PROC:
            source = MPI_PROC_NULL
        cdef _p_msg_p2p rmsg = message_p2p_recv(buf, source)
        cdef Request request = <Request>New(Request)
        with nogil: CHKERR( MPI_Imrecv_c(
            rmsg.buf, rmsg.count, rmsg.dtype,
            &message, &request.ob_mpi) )
        if self is not __MESSAGE_NO_PROC__:
            self.ob_mpi = message
        request.ob_buf = rmsg
        return request

    # Python Communication
    # --------------------
    #
    @classmethod
    def probe(
        cls,
        Comm comm: Comm,
        int source: int = ANY_SOURCE,
        int tag: int = ANY_TAG,
        Status status: Status | None = None,
    ) -> Self:
        """
        Blocking test for a matched message.
        """
        cdef Message message = <Message>New(cls)
        cdef MPI_Status *statusp = arg_Status(status)
        message.ob_buf = PyMPI_mprobe(source, tag, comm.ob_mpi,
                                      &message.ob_mpi, statusp)
        return message

    @classmethod
    def iprobe(
        cls,
        Comm comm: Comm,
        int source: int = ANY_SOURCE,
        int tag: int = ANY_TAG,
        Status status: Status | None = None,
    ) -> Self | None:
        """
        Nonblocking test for a matched message.
        """
        cdef int flag = 0
        cdef Message message = <Message>New(cls)
        cdef MPI_Status *statusp = arg_Status(status)
        message.ob_buf = PyMPI_improbe(source, tag, comm.ob_mpi, &flag,
                                       &message.ob_mpi, statusp)
        if flag == 0: return None
        return message

    def recv(self, Status status: Status | None = None) -> Any:
        """
        Blocking receive of matched message.
        """
        cdef object rmsg = self.ob_buf
        cdef MPI_Message message = self.ob_mpi
        cdef MPI_Status *statusp = arg_Status(status)
        rmsg = PyMPI_mrecv(rmsg, &message, statusp)
        if self is not __MESSAGE_NO_PROC__: self.ob_mpi = message
        if self.ob_mpi == MPI_MESSAGE_NULL: self.ob_buf = None
        return rmsg

    def irecv(self) -> Request:
        """
        Nonblocking receive of matched message.
        """
        cdef object rmsg = self.ob_buf
        cdef MPI_Message message = self.ob_mpi
        cdef Request request = <Request>New(Request)
        request.ob_buf = PyMPI_imrecv(rmsg, &message, &request.ob_mpi)
        if self is not __MESSAGE_NO_PROC__: self.ob_mpi = message
        if self.ob_mpi == MPI_MESSAGE_NULL: self.ob_buf = None
        return request

    # Integer Handle
    # --------------

    def toint(self) -> int:
        """
        Translate object to integer handle.
        """
        return MPI_Message_toint(self.ob_mpi)

    @classmethod
    def fromint(cls, arg: int, /) -> Message:
        """
        Translate integer handle to object.
        """
        return fromhandle(MPI_Message_fromint(arg))

    # Fortran Handle
    # --------------

    def py2f(self) -> int:
        """
        """
        return MPI_Message_c2f(self.ob_mpi)

    @classmethod
    def f2py(cls, arg: int) -> Message:
        """
        """
        return fromhandle(MPI_Message_f2c(arg))


cdef Message __MESSAGE_NULL__    = def_Message ( MPI_MESSAGE_NULL    , "MESSAGE_NULL"    )  # noqa
cdef Message __MESSAGE_NO_PROC__ = def_Message ( MPI_MESSAGE_NO_PROC , "MESSAGE_NO_PROC" )  # noqa


# Predefined message handles
# --------------------------

MESSAGE_NULL    = __MESSAGE_NULL__     #: Null message handle
MESSAGE_NO_PROC = __MESSAGE_NO_PROC__  #: No-proc message handle



# http://mvapich.cse.ohio-state.edu/benchmarks/

from mpi4py import MPI


def osu_bw(
    BENCHMARH="MPI Bandwidth Test",
    skip=10,
    loop=100,
    window_size=64,
    skip_large=2,
    loop_large=20,
    window_size_large=64,
    large_message_size=8192,
    MAX_MSG_SIZE=1 << 22,
):
    comm = MPI.COMM_WORLD
    myid = comm.Get_rank()
    numprocs = comm.Get_size()

    if numprocs != 2:
        if myid == 0:
            errmsg = "This test requires exactly two processes"
        else:
            errmsg = None
        raise SystemExit(errmsg)

    s_buf = allocate(MAX_MSG_SIZE)
    r_buf = allocate(MAX_MSG_SIZE)

    if myid == 0:
        print(f"# {BENCHMARH}")
    if myid == 0:
        print(f"# {'Size [B]':<8s}{'Latency [us]':>20s}")

    message_sizes = [2**i for i in range(30)]
    for size in message_sizes:
        if size > MAX_MSG_SIZE:
            break
        if size > large_message_size:
            skip = skip_large
            loop = loop_large
            window_size = window_size_large

        iterations = list(range(loop + skip))
        window_sizes = list(range(window_size))
        requests = [MPI.REQUEST_NULL] * window_size
        #
        comm.Barrier()
        if myid == 0:
            s_msg = [s_buf, size, MPI.BYTE]
            r_msg = [r_buf, 4, MPI.BYTE]
            for i in iterations:
                if i == skip:
                    t_start = MPI.Wtime()
                for j in window_sizes:
                    requests[j] = comm.Isend(s_msg, 1, 100)
                MPI.Request.Waitall(requests)
                comm.Recv(r_msg, 1, 101)
            t_end = MPI.Wtime()
        elif myid == 1:
            s_msg = [s_buf, 4, MPI.BYTE]
            r_msg = [r_buf, size, MPI.BYTE]
            for i in iterations:
                if i == skip:
                    pass
                for j in window_sizes:
                    requests[j] = comm.Irecv(r_msg, 0, 100)
                MPI.Request.Waitall(requests)
                comm.Send(s_msg, 0, 101)
        #
        if myid == 0:
            MB = size / 1e6 * loop * window_size
            s = t_end - t_start
            bandwidth = MB / s
            print(f"{size:-10d}{bandwidth:20.2f}")


def allocate(n):
    try:
        import mmap

        return mmap.mmap(-1, n)
    except (ImportError, OSError):
        try:
            from numpy import zeros

            return zeros(n, "B")
        except ImportError:
            from array import array

            return array("B", [0]) * n


if __name__ == "__main__":
    osu_bw()



# mpiexec -n 2 python ex-2.01.py

# Process 0 sends a message to process 1

# --------------------------------------------------------------------

import array

from mpi4py import MPI

if MPI.COMM_WORLD.Get_size() < 2:
    raise SystemExit from None

# --------------------------------------------------------------------

s = b"Hello there"

msg = array.array("B", b"\0" * 20)
tag = 99
status = MPI.Status()

myrank = MPI.COMM_WORLD.Get_rank()

if myrank == 0:
    msg[: len(s)] = array.array("B", s)
    MPI.COMM_WORLD.Send([msg, len(s) + 1, MPI.CHAR], 1, tag)
elif myrank == 1:
    MPI.COMM_WORLD.Recv([msg, 20, MPI.CHAR], 0, tag, status)

# --------------------------------------------------------------------

if myrank == 1:
    assert list(msg[: len(s)]) == list(s)
    assert msg[len(s)] == 0
    assert status.source == 0
    assert status.tag == tag
    assert status.error == MPI.SUCCESS
    assert status.Get_count(MPI.CHAR) == len(s) + 1

# --------------------------------------------------------------------



from seq import Seq

from mpi4py import MPI


def test():
    size = MPI.COMM_WORLD.Get_size()
    rank = MPI.COMM_WORLD.Get_rank()
    name = MPI.Get_processor_name()
    with Seq(MPI.COMM_WORLD, 1, 10):
        print(
            f"Hello, World! I am process {rank} of {size} on {name}.",
            flush=True,
        )


if __name__ == "__main__":
    test()



import os
import pathlib
import sys

wrkdir = pathlib.Path(__file__).resolve().parent
topdir = wrkdir.parent.parent
srcdir = topdir / "src"
cfgdir = topdir / "conf"

sys.path.insert(0, os.fspath(cfgdir))
from mpiapigen import Generator  # noqa: E402

generator = Generator()
libmpi_pxd = srcdir / "mpi4py" / "libmpi.pxd"
generator.parse_file(libmpi_pxd)
libmpi_h = wrkdir / "libmpi.h"
generator.dump_header_h(libmpi_h)

# from io import StringIO
# libmpi_h = StringIO()
# generator.dump_header_h(libmpi_h)
# print libmpi_h.read()

libmpi_c = wrkdir / "libmpi.c.in"
with libmpi_c.open("w", encoding="utf-8") as f:
    f.write(f"""\
#include <mpi.h>
#include "{srcdir}/lib-mpi.h"
""")



# -----------------------------------------------------------------------------

cdef inline int AddTypeMap(dict TD, const char tc[], Datatype dt) except -1:
    if dt.ob_mpi != MPI_DATATYPE_NULL:
        TD[pystr(tc)] = dt
        return 1
    return 0

# -----------------------------------------------------------------------------

cdef dict TypeDict = { }
_typedict = TypeDict

# boolean (C++)
AddTypeMap(TypeDict, "?" , __CXX_BOOL__ )  # PEP-3118 & NumPy
# boolean (C99)
AddTypeMap(TypeDict, "?" , __C_BOOL__   )  # PEP-3118 & NumPy
# (signed) integer
AddTypeMap(TypeDict, "b" , __SIGNED_CHAR__ )  # MPI-2
AddTypeMap(TypeDict, "h" , __SHORT__       )
AddTypeMap(TypeDict, "i" , __INT__         )
AddTypeMap(TypeDict, "l" , __LONG__        )
AddTypeMap(TypeDict, "q" , __LONG_LONG__   )
# unsigned integer
AddTypeMap(TypeDict, "B" , __UNSIGNED_CHAR__      )
AddTypeMap(TypeDict, "H" , __UNSIGNED_SHORT__     )
AddTypeMap(TypeDict, "I" , __UNSIGNED__           )
AddTypeMap(TypeDict, "L" , __UNSIGNED_LONG__      )
AddTypeMap(TypeDict, "Q" , __UNSIGNED_LONG_LONG__ )
# real floating
AddTypeMap(TypeDict, "e" , __REAL2__       )
AddTypeMap(TypeDict, "e" , __FLOAT16_T__   )
AddTypeMap(TypeDict, "f" , __FLOAT__       )
AddTypeMap(TypeDict, "d" , __DOUBLE__      )
AddTypeMap(TypeDict, "g" , __LONG_DOUBLE__ )  # PEP-3118 & NumPy
# complex floating (F90)
AddTypeMap(TypeDict, "Ze" , __COMPLEX4__       )
AddTypeMap(TypeDict, "Zf" , __COMPLEX8__       )  # PEP-3118
AddTypeMap(TypeDict, "Zd" , __COMPLEX16__      )  # PEP-3118
AddTypeMap(TypeDict, "E"  , __COMPLEX4__       )
AddTypeMap(TypeDict, "F"  , __COMPLEX8__       )  # NumPy
AddTypeMap(TypeDict, "D"  , __COMPLEX16__      )  # NumPy
# complex floating (C++)
AddTypeMap(TypeDict, "Zf" , __CXX_FLOAT_COMPLEX__       )  # PEP-3118
AddTypeMap(TypeDict, "Zd" , __CXX_DOUBLE_COMPLEX__      )  # PEP-3118
AddTypeMap(TypeDict, "Zg" , __CXX_LONG_DOUBLE_COMPLEX__ )  # PEP-3118
AddTypeMap(TypeDict, "F"  , __CXX_FLOAT_COMPLEX__       )  # NumPy
AddTypeMap(TypeDict, "D"  , __CXX_DOUBLE_COMPLEX__      )  # NumPy
AddTypeMap(TypeDict, "G"  , __CXX_LONG_DOUBLE_COMPLEX__ )  # NumPy
# complex floating (C99)
AddTypeMap(TypeDict, "Zf" , __C_FLOAT_COMPLEX__         )  # PEP-3118
AddTypeMap(TypeDict, "Zd" , __C_DOUBLE_COMPLEX__        )  # PEP-3118
AddTypeMap(TypeDict, "Zg" , __C_LONG_DOUBLE_COMPLEX__   )  # PEP-3118
AddTypeMap(TypeDict, "F"  , __C_FLOAT_COMPLEX__         )  # NumPy
AddTypeMap(TypeDict, "D"  , __C_DOUBLE_COMPLEX__        )  # NumPy
AddTypeMap(TypeDict, "G"  , __C_LONG_DOUBLE_COMPLEX__   )  # NumPy

# integer (F90)
AddTypeMap(TypeDict, "i1"  , __INTEGER1__  )  # NumPy
AddTypeMap(TypeDict, "i2"  , __INTEGER2__  )  # NumPy
AddTypeMap(TypeDict, "i4"  , __INTEGER4__  )  # NumPy
AddTypeMap(TypeDict, "i8"  , __INTEGER8__  )  # NumPy
AddTypeMap(TypeDict, "i16" , __INTEGER16__ )
# real floating (F90)
AddTypeMap(TypeDict, "f2"  , __REAL2__     )  # NumPy
AddTypeMap(TypeDict, "f4"  , __REAL4__     )  # NumPy
AddTypeMap(TypeDict, "f8"  , __REAL8__     )  # NumPy
AddTypeMap(TypeDict, "f16" , __REAL16__    )  # NumPy
# complex floating (F90)
AddTypeMap(TypeDict, "c4"  , __COMPLEX4__  )  # NumPy
AddTypeMap(TypeDict, "c8"  , __COMPLEX8__  )  # NumPy
AddTypeMap(TypeDict, "c16" , __COMPLEX16__ )  # NumPy
AddTypeMap(TypeDict, "c32" , __COMPLEX32__ )  # NumPy
# boolean (C99/C++)
AddTypeMap(TypeDict, "b1" , __CXX_BOOL__ )  # NumPy
AddTypeMap(TypeDict, "b1" , __C_BOOL__   )  # NumPy
# signed and unsigned integer (C)
if sizeof(char) == 1:
    AddTypeMap(TypeDict, "i1" ,  __SIGNED_CHAR__        )  # NumPy
    AddTypeMap(TypeDict, "u1" ,  __UNSIGNED_CHAR__      )  # NumPy
if sizeof(short) == 2:
    AddTypeMap(TypeDict, "i2" ,  __SHORT__              )  # NumPy
    AddTypeMap(TypeDict, "u2" ,  __UNSIGNED_SHORT__     )  # NumPy
if sizeof(long) == 4:
    AddTypeMap(TypeDict, "i4" ,  __LONG__               )  # NumPy  # ~> 32bit
    AddTypeMap(TypeDict, "u4" ,  __UNSIGNED_LONG__      )  # NumPy  # ~> 32bit
if sizeof(int) == 4:
    AddTypeMap(TypeDict, "i4" ,  __INT__                )  # NumPy
    AddTypeMap(TypeDict, "u4" ,  __UNSIGNED__           )  # NumPy
if sizeof(long long) == 8:
    AddTypeMap(TypeDict, "i8" ,  __LONG_LONG__          )  # NumPy
    AddTypeMap(TypeDict, "u8" ,  __UNSIGNED_LONG_LONG__ )  # NumPy
if sizeof(long) == 8:
    AddTypeMap(TypeDict, "i8" ,  __LONG__               )  # NumPy
    AddTypeMap(TypeDict, "u8" ,  __UNSIGNED_LONG__      )  # NumPy
# signed integer (C99)
AddTypeMap(TypeDict, "i1" , __INT8_T__   )  # NumPy
AddTypeMap(TypeDict, "i2" , __INT16_T__  )  # NumPy
AddTypeMap(TypeDict, "i4" , __INT32_T__  )  # NumPy
AddTypeMap(TypeDict, "i8" , __INT64_T__  )  # NumPy
# unsigned integer (C99)
AddTypeMap(TypeDict, "u1" , __UINT8_T__  )  # NumPy
AddTypeMap(TypeDict, "u2" , __UINT16_T__ )  # NumPy
AddTypeMap(TypeDict, "u4" , __UINT32_T__ )  # NumPy
AddTypeMap(TypeDict, "u8" , __UINT64_T__ )  # NumPy
# real (C) and complex (C99/C++) floating
AddTypeMap(TypeDict, "f2" , __FLOAT16_T__ )  # NumPy
if sizeof(float) == 4:
    AddTypeMap(TypeDict, "f4"  , __FLOAT__              )  # NumPy
    AddTypeMap(TypeDict, "c8"  , __CXX_FLOAT_COMPLEX__  )  # NumPy
    AddTypeMap(TypeDict, "c8"  , __C_FLOAT_COMPLEX__    )  # NumPy
if sizeof(double) == 8:
    AddTypeMap(TypeDict, "f8"  , __DOUBLE__             )  # NumPy
    AddTypeMap(TypeDict, "c16" , __CXX_DOUBLE_COMPLEX__ )  # NumPy
    AddTypeMap(TypeDict, "c16" , __C_DOUBLE_COMPLEX__   )  # NumPy
if sizeof(long double) == 12:
    AddTypeMap(TypeDict, "f12" , __LONG_DOUBLE__             )  # ~> NumPy i386
    AddTypeMap(TypeDict, "c24" , __CXX_LONG_DOUBLE_COMPLEX__ )  # ~> NumPy i386
    AddTypeMap(TypeDict, "c24" , __C_LONG_DOUBLE_COMPLEX__   )  # ~> NumPy i386
if sizeof(long double) == 16:
    AddTypeMap(TypeDict, "f16" , __LONG_DOUBLE__             )  # NumPy
    AddTypeMap(TypeDict, "c32" , __CXX_LONG_DOUBLE_COMPLEX__ )  # NumPy
    AddTypeMap(TypeDict, "c32" , __C_LONG_DOUBLE_COMPLEX__   )  # NumPy

# ssize_t and size_t (C)
if sizeof(size_t) == sizeof(long long):
    AddTypeMap(TypeDict, "n" , __LONG_LONG__          )
    AddTypeMap(TypeDict, "N" , __UNSIGNED_LONG_LONG__ )
if sizeof(size_t) == sizeof(long):
    AddTypeMap(TypeDict, "n" , __LONG__               )
    AddTypeMap(TypeDict, "N" , __UNSIGNED_LONG__      )
if sizeof(size_t) == sizeof(int):
    AddTypeMap(TypeDict, "n" , __INT__                )  # ~> 32bit
    AddTypeMap(TypeDict, "N" , __UNSIGNED__           )  # ~> 32bit
if sizeof(size_t) == sizeof(MPI_Count):
    AddTypeMap(TypeDict, "n" , __COUNT__              )

# intptr_t and uintptr_t (C99)
if sizeof(Py_intptr_t) == sizeof(long long):
    AddTypeMap(TypeDict, "p" , __LONG_LONG__          )  # NumPy
    AddTypeMap(TypeDict, "P" , __UNSIGNED_LONG_LONG__ )  # NumPy
if sizeof(Py_intptr_t) == sizeof(long):
    AddTypeMap(TypeDict, "p" , __LONG__               )  # NumPy
    AddTypeMap(TypeDict, "P" , __UNSIGNED_LONG__      )  # NumPy
if sizeof(Py_intptr_t) == sizeof(int):
    AddTypeMap(TypeDict, "p" , __INT__                )  # ~> NumPy 32bit
    AddTypeMap(TypeDict, "P" , __UNSIGNED__           )  # ~> NumPy 32bit
if sizeof(Py_intptr_t) == sizeof(MPI_Aint):
    AddTypeMap(TypeDict, "p" , __AINT__               )  # NumPy

# character
AddTypeMap(TypeDict, "c" , __CHAR__ )  # PEP-3118 & NumPy
AddTypeMap(TypeDict, "a" , __CHAR__ )  # NumPy
AddTypeMap(TypeDict, "S" , __CHAR__ )  # NumPy
AddTypeMap(TypeDict, "S1", __CHAR__ )  # NumPy
AddTypeMap(TypeDict, "s",  __CHAR__ )  # PEP-3118
AddTypeMap(TypeDict, "1s", __CHAR__ )  # PEP-3118
# wide character
if sizeof(wchar_t) == 4:
    AddTypeMap(TypeDict, "U" , __WCHAR__          )  # NumPy
    AddTypeMap(TypeDict, "U1", __WCHAR__          )  # NumPy
# UTF-16/UCS-2
if sizeof(short) == 2:
    AddTypeMap(TypeDict, "u" , __UNSIGNED_SHORT__ )  # PEP-3118
    AddTypeMap(TypeDict, "1u", __UNSIGNED_SHORT__ )  # PEP-3118
if 2 == 2:
    AddTypeMap(TypeDict, "u" , __UINT16_T__       )  # PEP-3118
    AddTypeMap(TypeDict, "1u", __UINT16_T__       )  # PEP-3118
if sizeof(wchar_t) == 2:
    AddTypeMap(TypeDict, "u" , __WCHAR__          )  # PEP-3118  # ~> win32
    AddTypeMap(TypeDict, "1u", __WCHAR__          )  # PEP-3118  # ~> win32
# UTF-32/UCS-4
if sizeof(int) == 4:
    AddTypeMap(TypeDict, "w" , __UNSIGNED__       )  # PEP-3118
    AddTypeMap(TypeDict, "1w", __UNSIGNED__       )  # PEP-3118
if 4 == 4:
    AddTypeMap(TypeDict, "w" , __UINT32_T__       )  # PEP-3118
    AddTypeMap(TypeDict, "1w", __UINT32_T__       )  # PEP-3118
if sizeof(wchar_t) == 4:
    AddTypeMap(TypeDict, "w" , __WCHAR__          )  # PEP-3118
    AddTypeMap(TypeDict, "1w", __WCHAR__          )  # PEP-3118
# datetime
AddTypeMap(TypeDict, "M" , __INT64_T__ )  # NumPy
AddTypeMap(TypeDict, "M8", __INT64_T__ )  # NumPy
# timedelta
AddTypeMap(TypeDict, "m" , __INT64_T__ )  # NumPy
AddTypeMap(TypeDict, "m8", __INT64_T__ )  # NumPy

# -----------------------------------------------------------------------------

cdef dict CTypeDict = { }
_typedict_c = CTypeDict

AddTypeMap(CTypeDict, "?" , __C_BOOL__ )

AddTypeMap(CTypeDict, "b" , __SIGNED_CHAR__ )
AddTypeMap(CTypeDict, "h" , __SHORT__       )
AddTypeMap(CTypeDict, "i" , __INT__         )
AddTypeMap(CTypeDict, "l" , __LONG__        )
AddTypeMap(CTypeDict, "q" , __LONG_LONG__   )

AddTypeMap(CTypeDict, "B" , __UNSIGNED_CHAR__      )
AddTypeMap(CTypeDict, "H" , __UNSIGNED_SHORT__     )
AddTypeMap(CTypeDict, "I" , __UNSIGNED__           )
AddTypeMap(CTypeDict, "L" , __UNSIGNED_LONG__      )
AddTypeMap(CTypeDict, "Q" , __UNSIGNED_LONG_LONG__ )

AddTypeMap(CTypeDict, "f" , __FLOAT__       )
AddTypeMap(CTypeDict, "d" , __DOUBLE__      )
AddTypeMap(CTypeDict, "g" , __LONG_DOUBLE__ )

AddTypeMap(CTypeDict, "F" , __C_FLOAT_COMPLEX__       )
AddTypeMap(CTypeDict, "D" , __C_DOUBLE_COMPLEX__      )
AddTypeMap(CTypeDict, "G" , __C_LONG_DOUBLE_COMPLEX__ )

AddTypeMap(CTypeDict, "b1" , __C_BOOL__   )

AddTypeMap(CTypeDict, "i1" , __INT8_T__   )
AddTypeMap(CTypeDict, "i2" , __INT16_T__  )
AddTypeMap(CTypeDict, "i4" , __INT32_T__  )
AddTypeMap(CTypeDict, "i8" , __INT64_T__  )

AddTypeMap(CTypeDict, "u1" , __UINT8_T__  )
AddTypeMap(CTypeDict, "u2" , __UINT16_T__ )
AddTypeMap(CTypeDict, "u4" , __UINT32_T__ )
AddTypeMap(CTypeDict, "u8" , __UINT64_T__ )

AddTypeMap(CTypeDict, "f2" , __FLOAT16_T__ )
if sizeof(float) == 4:
    AddTypeMap(CTypeDict, "f4"  , __FLOAT__            )
    AddTypeMap(CTypeDict, "c8"  , __C_FLOAT_COMPLEX__  )
if sizeof(double) == 8:
    AddTypeMap(CTypeDict, "f8"  , __DOUBLE__           )
    AddTypeMap(CTypeDict, "c16" , __C_DOUBLE_COMPLEX__ )
if sizeof(long double) == 12:
    AddTypeMap(CTypeDict, "f12" , __LONG_DOUBLE__           )  # ~> i386
    AddTypeMap(CTypeDict, "c24" , __C_LONG_DOUBLE_COMPLEX__ )  # ~> i386
if sizeof(long double) == 16:
    AddTypeMap(CTypeDict, "f16" , __LONG_DOUBLE__           )
    AddTypeMap(CTypeDict, "c32" , __C_LONG_DOUBLE_COMPLEX__ )

# -----------------------------------------------------------------------------

cdef dict FTypeDict = { }
_typedict_f = FTypeDict

AddTypeMap(FTypeDict, "l"   , __LOGICAL__          )
AddTypeMap(FTypeDict, "i"   , __INTEGER__          )
AddTypeMap(FTypeDict, "r"   , __REAL__             )
AddTypeMap(FTypeDict, "s"   , __REAL__             )
AddTypeMap(FTypeDict, "d"   , __DOUBLE_PRECISION__ )
AddTypeMap(FTypeDict, "c"   , __COMPLEX__          )
AddTypeMap(FTypeDict, "z"   , __DOUBLE_COMPLEX__   )

AddTypeMap(FTypeDict, "l1"  , __LOGICAL1__  )
AddTypeMap(FTypeDict, "l2"  , __LOGICAL2__  )
AddTypeMap(FTypeDict, "l4"  , __LOGICAL4__  )
AddTypeMap(FTypeDict, "l8"  , __LOGICAL8__  )
AddTypeMap(FTypeDict, "l16" , __LOGICAL16__ )

AddTypeMap(FTypeDict, "i1"  , __INTEGER1__  )
AddTypeMap(FTypeDict, "i2"  , __INTEGER2__  )
AddTypeMap(FTypeDict, "i4"  , __INTEGER4__  )
AddTypeMap(FTypeDict, "i8"  , __INTEGER8__  )
AddTypeMap(FTypeDict, "i16" , __INTEGER16__ )

AddTypeMap(FTypeDict, "r2"  , __REAL2__     )
AddTypeMap(FTypeDict, "r4"  , __REAL4__     )
AddTypeMap(FTypeDict, "r8"  , __REAL8__     )
AddTypeMap(FTypeDict, "r16" , __REAL16__    )

AddTypeMap(FTypeDict, "c4"  , __COMPLEX4__  )
AddTypeMap(FTypeDict, "c8"  , __COMPLEX8__  )
AddTypeMap(FTypeDict, "c16" , __COMPLEX16__ )
AddTypeMap(FTypeDict, "c32" , __COMPLEX32__ )

# -----------------------------------------------------------------------------



import os
import pathlib
import unittest
import warnings

import mpi4py


class TestRC(unittest.TestCase):
    #
    @staticmethod
    def newrc():
        rc = type(mpi4py.rc)()
        rc(initialize=rc.initialize)
        rc(threads=rc.threads)
        rc(thread_level=rc.thread_level)
        rc(finalize=rc.finalize)
        rc(fast_reduce=rc.fast_reduce)
        rc(recv_mprobe=rc.recv_mprobe)
        rc(irecv_bufsz=rc.irecv_bufsz)
        rc(errors=rc.errors)
        return rc

    def testCallKwArgs(self):
        rc = self.newrc()
        kwargs = rc.__dict__.copy()
        rc(**kwargs)

    def testInitKwArgs(self):
        rc = self.newrc()
        kwargs = rc.__dict__.copy()
        rc = type(mpi4py.rc)(**kwargs)

    def testBadAttribute(self):
        with self.assertRaises(TypeError):
            mpi4py.rc(ABCXYZ=123456)
        with self.assertRaises(TypeError):
            mpi4py.rc.ABCXYZ = 123456
        with self.assertRaises(AttributeError):
            _ = mpi4py.rc.ABCXYZ

    def testRepr(self):
        repr(mpi4py.rc)


class TestConfig(unittest.TestCase):
    #
    def testGetInclude(self):
        include = mpi4py.get_include()
        self.assertIsInstance(include, str)
        include = pathlib.Path(include)
        self.assertTrue(include.is_dir())
        header = include / "mpi4py" / "mpi4py.h"
        self.assertTrue(header.is_file())

    def testGetConfig(self):
        conf = mpi4py.get_config()
        self.assertIsInstance(conf, dict)
        mpicc = conf.get("mpicc")
        if mpicc is not None:
            self.assertTrue(pathlib.Path(mpicc).exists())


class TestProfile(unittest.TestCase):
    #
    def testProfile(self):
        import struct
        import sysconfig

        bits = struct.calcsize("P") * 8
        triplet = sysconfig.get_config_var("MULTIARCH") or ""
        libpath = [
            f"{prefix}{suffix}"
            for prefix in ("/lib", "/usr/lib")
            for suffix in (bits, f"/{triplet}", "")
        ]
        fspath = (os.fsencode, os.fsdecode, pathlib.Path)
        libraries = (
            "c",
            "libc.so.6",
            "m",
            "libm.so.6",
            "dl",
            "libdl.so.2",
        )

        def mpi4py_profile(*args, **kwargs):
            try:
                mpi4py.profile(*args, **kwargs)
            except ValueError:
                pass

        if os.name != "posix":
            with warnings.catch_warnings():
                warnings.simplefilter("error")
                with self.assertRaises(UserWarning):
                    mpi4py.profile(struct.__file__)
            return
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for libname in libraries:
                mpi4py_profile(libname, path=libpath)
                for fs in fspath:
                    mpi4py_profile(libname, path=map(fs, libpath))
                for path in libpath:
                    mpi4py_profile(libname, path=path)
                    for fsp in fspath:
                        mpi4py_profile(libname, path=fsp(path))
            warnings.simplefilter("error")
            with self.assertRaises(UserWarning):
                mpi4py.profile("hosts", path=["/etc"])
            with self.assertRaises(ValueError):
                mpi4py.profile("@querty")
            with self.assertRaises(ValueError):
                mpi4py.profile("@querty", path="/usr/lib")
            with self.assertRaises(ValueError):
                mpi4py.profile("@querty", path=["/usr/lib"])
            with self.assertRaises(ValueError):
                mpi4py.profile("@querty")


if __name__ == "__main__":
    unittest.main()



import itertools
import os
import pathlib
import sys

from mpi4py import MPI
from mpi4py.util.dtlib import (
    from_numpy_dtype as fromnumpy,
    to_numpy_dtype as tonumpy,
)

try:
    import mpitestutil as testutil
    import mpiunittest as unittest
except ImportError:
    sys.path.append(os.fspath(pathlib.Path(__file__).resolve().parent))
    import mpitestutil as testutil
    import mpiunittest as unittest

try:
    import numpy

    np_dtype = numpy.dtype
    np_version = tuple(map(int, numpy.__version__.split(".", 2)[:2]))
except ImportError:
    numpy = None
    np_dtype = None
    np_version = None

typecodes = list("?cbhilqpBHILQPefdgFDG")
typecodes += [f"b{n:d}" for n in (1,)]
typecodes += [f"i{n:d}" for n in (1, 2, 4, 8)]
typecodes += [f"u{n:d}" for n in (1, 2, 4, 8)]
typecodes += [f"f{n:d}" for n in (2, 4, 8)]
if os.environ.get("COVERAGE_RUN") == "true":
    typecodes = [*list("cif"), "b1", "i8", "f8"]

if np_version and np_version < (1, 17):
    for tc in "LFDG":
        if tc in typecodes:
            typecodes.remove(tc)

if (
    MPI.FLOAT16_T == MPI.DATATYPE_NULL
    or unittest.is_mpi("mpich(<5.0.0)")
    or unittest.is_mpi("openmpi(<5.0.0)")
    or unittest.is_mpi("impi")
):
    for tc in ("e", "f2"):
        if tc in typecodes:
            typecodes.remove(tc)

if unittest.is_mpi("mpich(<4.0.0)"):
    for tc in "FDG":
        if tc in typecodes:
            typecodes.remove(tc)

if unittest.is_mpi("impi(>=2021.12.0)") and os.name == "nt":
    for tc in (*"lLg", "i4", "u4"):
        if tc in typecodes:
            typecodes.remove(tc)

datatypes_c = [MPI.Datatype.fromcode(t) for t in typecodes] + [
    MPI.BYTE,
    MPI.AINT,
    MPI.OFFSET,
    MPI.COUNT,
]

mpipairtypes = [
    MPI.SHORT_INT,
    MPI.INT_INT,
    MPI.LONG_INT,
    MPI.FLOAT_INT,
    MPI.DOUBLE_INT,
    MPI.LONG_DOUBLE_INT,
]

mpif77types = [
    MPI.CHARACTER,
    MPI.LOGICAL,
    MPI.INTEGER,
    MPI.REAL,
    MPI.DOUBLE_PRECISION,
    MPI.COMPLEX,
    MPI.DOUBLE_COMPLEX,
]

mpif90types = [
    MPI.LOGICAL1,
    MPI.LOGICAL2,
    MPI.LOGICAL4,
    MPI.LOGICAL8,
    MPI.LOGICAL16,
    MPI.INTEGER1,
    MPI.INTEGER2,
    MPI.INTEGER4,
    MPI.INTEGER8,
    MPI.INTEGER16,
    MPI.REAL2,
    MPI.REAL4,
    MPI.REAL8,
    MPI.REAL16,
    MPI.COMPLEX4,
    MPI.COMPLEX8,
    MPI.COMPLEX16,
    MPI.COMPLEX32,
]

for typelist in [
    datatypes_c,
    mpif77types,
    mpif90types,
]:
    typelist[:] = [t for t in typelist if testutil.has_datatype(t)]
del typelist

datatypes = []
datatypes += datatypes_c
datatypes += mpif77types
datatypes += mpif90types


def try_dtype(*args):
    for spec in args:
        if isinstance(spec, MPI.Datatype):
            spec = spec.typestr
        if np_dtype is not None:
            try:
                np_dtype(spec)
            except TypeError:
                return False
    return True


class TestUtilDTLib(unittest.TestCase):
    #
    def check(self, arg, *args):
        if numpy is None:
            if isinstance(arg, MPI.Datatype):
                mt1 = arg.Dup()
                dt1 = tonumpy(mt1)
                mt1.Free()
            return
        if not try_dtype(arg):
            return
        if isinstance(arg, MPI.Datatype):
            mt1 = arg.Dup()
            dt1 = tonumpy(mt1)
        else:
            dt1 = np_dtype(arg, *args)
            mt1 = fromnumpy(dt1)
        dt2 = tonumpy(mt1)
        mt2 = fromnumpy(dt2)
        dt3 = tonumpy(mt2)
        mt3 = fromnumpy(dt3)
        try:
            self.assertEqual(dt1, dt2)
            self.assertEqual(dt2, dt3)
            if isinstance(arg, MPI.Datatype):
                if arg.combiner not in (
                    MPI.COMBINER_INDEXED,
                    MPI.COMBINER_HINDEXED,
                    MPI.COMBINER_INDEXED_BLOCK,
                    MPI.COMBINER_HINDEXED_BLOCK,
                ):
                    self.assertEqual(dt1.itemsize, mt1.extent)
            self.assertEqual(dt2.itemsize, mt2.extent)
            self.assertEqual(dt3.itemsize, mt3.extent)
        finally:
            mt1.Free()
            mt2.Free()
            mt3.Free()

    def testBasic(self):
        for spec in typecodes:
            with self.subTest(spec=spec):
                self.check(spec)
        for mpit in datatypes:
            with self.subTest(name=mpit.name):
                self.check(mpit)

    def testSubarray1(self):
        shapes = [
            (1,),
            (1, 1),
            (1, 1, 1),
            (3,),
            (3, 4),
            (2, 3, 4),
        ]
        for dt, shape in itertools.product(typecodes, shapes):
            spec = f"{shape}{dt}"
            with self.subTest(spec=spec):
                self.check(spec)

    def testSubarray2(self):
        shapes = [
            (1,),
            (1, 1),
            (1, 1, 1),
            (3,),
            (3, 4),
            (2, 3, 4),
        ]
        orders = [MPI.ORDER_C, MPI.ORDER_FORTRAN]
        for mt, shape, order in itertools.product(datatypes, shapes, orders):
            if not try_dtype(mt):
                continue
            with self.subTest(name=mt.name, shape=shape, order=order):
                starts = (0,) * len(shape)
                mt1 = mt.Create_subarray(shape, shape, starts, order)
                self.check(mt1)
                mt1.Free()

    @unittest.skipMPI("msmpi")
    def testStruct1(self):
        shapes = [
            (),
            (1,),
            (3,),
            (3, 5),
        ]
        iter1 = itertools.product(shapes, typecodes)
        iter2 = itertools.product(shapes, typecodes)
        iterN = itertools.product(iter1, iter2)
        iterA = iter([False, True])
        for nt, align in itertools.product(iterN, iterA):
            s1, t1, s2, t2 = sum(nt, ())
            spec = f"{s1}{t1},{s2}{t2}"
            with self.subTest(spec=spec, align=align):
                self.check(spec, align)

    @unittest.skipMPI("msmpi")
    def testStruct2(self):
        iter1 = iter(typecodes)
        iter2 = iter(typecodes)
        iter3 = iter(typecodes)
        iterN = itertools.product(iter1, iter2, iter3)
        iterA = iter([False, True])
        for tp, align in itertools.product(iterN, iterA):
            t1, t2, t3 = tp
            spec = f"{t1},{t2},{t3}"
            with self.subTest(spec=spec, align=align):
                self.check(spec, align)

    @unittest.skipMPI("msmpi")
    def testStruct3(self):
        blens = [1, 2, 3]
        disps = [1, 27, 71]
        types = [MPI.INT, MPI.DOUBLE, MPI.INT]
        mt1 = MPI.Datatype.Create_struct(blens, disps, types)
        mt2 = MPI.Datatype.Create_struct([1], [0], [mt1])
        self.check(mt1)
        self.check(mt2)
        mt1.Free()
        mt2.Free()

    def makeStruct(self, dt, mt):
        dt = numpy.dtype(dt).str
        stp = numpy.dtype(f"B,{dt},B", align=True)

        def off(i):
            return stp.fields[stp.names[i]][1]

        blens = [1, 1, 1]
        disps = [0, off(1), off(2)]
        types = [MPI.BYTE, mt, MPI.BYTE]
        mtp = MPI.Datatype.Create_struct(blens, disps, types)
        return stp, mtp

    @unittest.skipMPI("msmpi")
    @unittest.skipIf(numpy is None, "numpy")
    def testStruct4(self):
        for t in typecodes:
            with self.subTest(typecode=t):
                dt0 = np_dtype(t)
                mt0 = fromnumpy(dt0)
                stp, mt1 = self.makeStruct(t, mt0)
                ex1 = stp.itemsize
                for n, mt in (
                    (1, mt1),
                    (1, mt1.Dup()),
                    (1, mt1.Create_resized(0, 1 * ex1)),
                    (3, mt1.Create_resized(0, 3 * ex1)),
                    (3, mt1.Create_contiguous(3)),
                    (5, mt1.Create_subarray([5], [5], [0])),
                    (7, MPI.Datatype.Create_struct([7], [0], [mt1])),
                ):
                    dt = tonumpy(mt)
                    self.assertEqual(mt.extent, n * ex1)
                    self.assertEqual(dt.itemsize, n * ex1)
                    self.assertTrue(dt.isalignedstruct)
                    self.check(mt)
                    self.check(dt)
                    if mt != mt1:
                        mt.Free()
                mt0.Free()
                mt1.Free()

    @unittest.skipMPI("msmpi")
    @unittest.skipIf(numpy is None, "numpy")
    def testStruct5(self):
        for t1, t2 in itertools.product(*[typecodes] * 2):
            with self.subTest(t1=t1, t2=t2):
                dtlist = []
                dt = np_dtype(f"c,{t1},{t2},c", align=True)
                dtlist.append(dt)
                for _ in range(3):
                    dt = np_dtype([("", dt)] * 2, align=True)
                    dtlist.append(dt)
                for dt in dtlist:
                    mt = fromnumpy(dt)
                    dt2 = tonumpy(mt)
                    mt.Free()
                    self.assertEqual(dt, dt2)

    def testVector(self):
        for mt in datatypes:
            if not try_dtype(mt):
                continue
            with self.subTest(name=mt.name):
                mt1 = mt.Create_vector(3, 4, 6)
                mt2 = mt.Create_hvector(3, 4, 6 * mt.extent)
                self.check(mt1)
                self.check(mt2)
                dt1 = tonumpy(mt1)
                dt2 = tonumpy(mt2)
                self.check(dt1)
                self.check(dt2)
                self.assertEqual(dt1, dt2)
                mt3 = mt1.Create_vector(2, 3, 4)
                mt4 = mt2.Create_hvector(2, 3, 4 * mt2.extent)
                self.check(mt3)
                self.check(mt4)
                dt3 = tonumpy(mt3)
                dt4 = tonumpy(mt4)
                self.check(dt3)
                self.check(dt4)
                self.assertEqual(dt3, dt4)
                mt3.Free()
                mt4.Free()
                mt1.Free()
                mt2.Free()

    def testHVector(self):
        for mt in datatypes:
            if not try_dtype(mt):
                continue
            with self.subTest(name=mt.name):
                mt1 = mt.Create_hvector(3, 4, 6 * mt.extent + 1)
                mt2 = mt1.Dup()
                self.check(mt1)
                self.check(mt2)
                dt1 = tonumpy(mt1)
                dt2 = tonumpy(mt2)
                self.check(dt1)
                self.check(dt2)
                self.assertEqual(dt1, dt2)
                mt3 = mt1.Create_hvector(2, 3, 4 * mt1.extent + 1)
                mt4 = mt2.Create_hvector(2, 3, 4 * mt2.extent + 1)
                self.check(mt3)
                self.check(mt4)
                dt3 = tonumpy(mt3)
                dt4 = tonumpy(mt4)
                self.check(dt3)
                self.check(dt4)
                self.assertEqual(dt3, dt4)
                mt3.Free()
                mt4.Free()
                mt1.Free()
                mt2.Free()

    def testIndexed(self):
        disps = [1, 6, 12]
        for mt in datatypes:
            if not try_dtype(mt):
                continue
            with self.subTest(name=mt.name):
                mt1 = mt.Create_indexed([4] * 3, disps)
                mt2 = mt.Create_indexed_block(4, disps)
                self.check(mt1)
                self.check(mt2)
                dt1 = tonumpy(mt1)
                dt2 = tonumpy(mt2)
                self.check(dt1)
                self.check(dt2)
                self.assertEqual(dt1, dt2)
                mt3 = mt1.Create_indexed([1], [0])
                mt4 = mt2.Create_indexed_block(1, [0])
                self.check(mt3)
                self.check(mt4)
                dt3 = tonumpy(mt3)
                dt4 = tonumpy(mt4)
                self.check(dt3)
                self.check(dt4)
                self.assertEqual(dt3, dt4)
                mt3.Free()
                mt4.Free()
                mt1.Free()
                mt2.Free()

    def testHIndexed(self):
        disps = [0, 6, 12]
        for mt in datatypes:
            if not try_dtype(mt):
                continue
            with self.subTest(name=mt.name):
                mt1 = mt.Create_hindexed(
                    [4] * 3, [d * mt.extent + 1 for d in disps]
                )
                mt2 = mt.Create_hindexed_block(
                    4, [d * mt.extent + 1 for d in disps]
                )
                self.check(mt1)
                self.check(mt2)
                dt1 = tonumpy(mt1)
                dt2 = tonumpy(mt2)
                self.check(dt1)
                self.check(dt2)
                self.assertEqual(dt1, dt2)
                mt3 = mt1.Create_hindexed([1], [0])
                mt4 = mt2.Create_hindexed_block(1, [0])
                self.check(mt3)
                self.check(mt4)
                mt3.Free()
                mt4.Free()
                mt1.Free()
                mt2.Free()

    @unittest.skipMPI("msmpi")
    def testF77(self):
        for mt in mpif77types:
            dt = tonumpy(mt)
            if np_dtype is not None:
                self.assertEqual(dt.itemsize, mt.extent)

    @unittest.skipMPI("msmpi")
    def testF90(self):
        for mt in mpif90types:
            if np_dtype is not None:
                typestr = mt.typestr
                try:
                    np_dtype(typestr)
                except TypeError:
                    continue
            dt = tonumpy(mt)
            if np_dtype is not None:
                self.assertEqual(dt.itemsize, mt.extent)

    @unittest.skipMPI("msmpi")
    def testF90Integer(self):
        try:
            mt = MPI.Datatype.Create_f90_integer(1)
        except NotImplementedError:
            mt = MPI.DATATYPE_NULL
        if not testutil.has_datatype(mt):
            self.skipTest("mpi-type-create-f90-integer")
        for r in range(1, 19):
            with self.subTest(r=r):
                mt = MPI.Datatype.Create_f90_integer(r)
                dt = tonumpy(mt)
                if np_dtype is not None:
                    self.assertEqual(dt.kind, "i")
                    self.assertEqual(dt.itemsize, mt.extent)
                    size = mt.Get_size()
                    tstr = f"i{size}"
                    stp, mtp = self.makeStruct(tstr, mt)
                    self.assertEqual(stp.itemsize, mtp.extent)
                    self.check(mtp)
                    mtp.Free()

    @unittest.skipMPI("msmpi")
    def testF90Real(self):
        try:
            mt = MPI.Datatype.Create_f90_real(7, MPI.UNDEFINED)
        except NotImplementedError:
            mt = MPI.DATATYPE_NULL
        if not testutil.has_datatype(mt):
            self.skipTest("mpi-type-create-f90-real")
        for p in (6, 7, 14, 15):
            with self.subTest(p=p):
                mt = MPI.Datatype.Create_f90_real(p, MPI.UNDEFINED)
                dt = tonumpy(mt)
                if np_dtype is not None:
                    self.assertEqual(dt.kind, "f")
                    self.assertEqual(dt.itemsize, mt.extent)
                    size = mt.Get_size()
                    tstr = f"i{size}"
                    stp, mtp = self.makeStruct(tstr, mt)
                    self.assertEqual(stp.itemsize, mtp.extent)
                    self.check(mtp)
                    mtp.Free()

    @unittest.skipMPI("msmpi")
    def testF90Complex(self):
        try:
            mt = MPI.Datatype.Create_f90_complex(7, MPI.UNDEFINED)
        except NotImplementedError:
            mt = MPI.DATATYPE_NULL
        if not testutil.has_datatype(mt):
            self.skipTest("mpi-type-create-f90-complex")
        for p in (6, 7, 14, 15):
            with self.subTest(p=p):
                mt = MPI.Datatype.Create_f90_complex(p, MPI.UNDEFINED)
                dt = tonumpy(mt)
                if np_dtype is not None:
                    self.assertEqual(dt.kind, "c")
                    self.assertEqual(dt.itemsize, mt.extent)

    def testPair(self):
        for mt in mpipairtypes:
            with self.subTest(datatype=mt.name):
                dt = tonumpy(mt)
                if np_dtype is not None:
                    self.assertTrue(dt.isalignedstruct)
                    self.assertEqual(dt.itemsize, mt.extent)
        integral = "bhilqpBHILQP"
        floating = "fdg"
        vtypes = integral + floating
        itypes = integral
        for vcode, icode in itertools.product(vtypes, itypes):
            value = MPI.Datatype.fromcode(vcode)
            index = MPI.Datatype.fromcode(icode)
            pair = MPI.Datatype.Get_value_index(value, index)
            if pair == MPI.DATATYPE_NULL:
                continue
            vt, it, pt = map(tonumpy, (value, index, pair))
            dt = (f"{vt},{it}", {"align": True})
            if np_dtype is not None:
                dt = np_dtype(dt[0], **dt[1])
            self.assertEqual(pt, dt)

    def testPairStruct(self):
        cases = [mpipairtypes] * 3 + [[False, True]]
        for mt1, mt2, mt3, dup in itertools.product(*cases):
            with self.subTest(mt1=mt1.name, mt2=mt2.name, mt3=mt3.name):
                if dup:
                    mt1 = mt1.Dup()
                    mt2 = mt2.Dup()
                    mt3 = mt3.Dup()
                align = max(mt.extent for mt in (mt1, mt2, mt3))
                structtype = MPI.Datatype.Create_struct(
                    [1, 1, 1],
                    [0, align, align * 2],
                    [mt1, mt2, mt3],
                )
                if dup:
                    mt1.Free()
                    mt2.Free()
                    mt3.Free()
                dt = tonumpy(structtype)
                structtype.Free()
                if np_dtype is not None:
                    self.assertTrue(dt.isalignedstruct)

    def testAlignmentComplex(self):
        complexcodes = list("FDG")
        complexcodes += [f"c{n}" for n in (8, 16)]
        for t in typecodes + complexcodes:
            with self.subTest(typecode=t):
                datatype = MPI.Datatype.fromcode(t)
                alignment1 = MPI._typealign(datatype)
                if np_dtype is not None:
                    alignment2 = np_dtype(t).alignment
                    self.assertEqual(alignment1, alignment2)

    def testAlignmentPair(self):
        for pairtype in mpipairtypes:
            alignment1 = MPI._typealign(pairtype)
            self.assertIn(alignment1, (2, 4, 8, 16))
            if np_dtype is not None:
                alignment2 = tonumpy(pairtype).alignment
                self.assertEqual(alignment1, alignment2)

    def testAlignmentStruct(self):
        off = MPI.DOUBLE.extent
        structtype = MPI.Datatype.Create_struct(
            [1, 1],
            [0, off],
            [MPI.INT, MPI.DOUBLE],
        )
        alignment = MPI._typealign(structtype)
        self.assertIsNone(alignment)
        structtype.Free()

    def testMissingNumPy(self):
        from mpi4py.util import dtlib

        np_dtype = dtlib._np_dtype
        dtlib._np_dtype = None
        try:
            for t in typecodes:
                with self.subTest(typecode=t):
                    mt = MPI.Datatype.fromcode(t)
                    dt = tonumpy(mt)
                    code = mt.tocode()
                    self.assertEqual(dt, code)
                    arraytype = mt.Create_contiguous(7)
                    dt = tonumpy(arraytype)
                    arraytype.Free()
                    self.assertIsInstance(dt, tuple)
                    self.assertEqual(dt[0], code)
                    self.assertEqual(dt[1], (7,))
                    structtype = MPI.Datatype.Create_struct(
                        [1, 1],
                        [0, mt.extent],
                        [mt, mt],
                    )
                    dt = tonumpy(structtype)
                    structtype.Free()
                    self.assertIsInstance(dt, dict)
                    self.assertEqual(dt["formats"], [code] * 2)
                    self.assertEqual(dt["offsets"], [0, mt.extent])
                    self.assertEqual(dt["itemsize"], mt.extent * 2)
                    self.assertTrue(dt["aligned"])
            with self.assertRaises(RuntimeError):
                fromnumpy(None)
        finally:
            dtlib._np_dtype = np_dtype

    @unittest.skipIf(numpy is None, "numpy")
    def testFailures(self):
        endian = ">" if np_dtype("<i").isnative else "<"
        self.assertRaises(ValueError, fromnumpy, np_dtype(endian + "i"))
        self.assertRaises(ValueError, fromnumpy, np_dtype("O"))
        self.assertRaises(ValueError, fromnumpy, np_dtype("V"))
        self.assertRaises(ValueError, tonumpy, MPI.DATATYPE_NULL)
        mt = MPI.INT.Create_resized(0, 32)
        self.assertRaises(ValueError, tonumpy, mt)
        mt.Free()
        mt = MPI.INT.Create_subarray([2], [1], [0])
        self.assertRaises(ValueError, tonumpy, mt)
        mt.Free()


if __name__ == "__main__":
    unittest.main()



from mpi4py import MPI


class Intracomm(MPI.Intracomm):
    """
    Intracommunicator class with scalable, point-to-point based
    implementations of global reduction operations.
    """

    def __new__(cls, comm=None):
        return super().__new__(cls, comm)

    def reduce(self, sendobj=None, recvobj=None, op=MPI.SUM, root=0):
        size = self.size
        rank = self.rank
        assert 0 <= root < size
        tag = MPI.COMM_WORLD.Get_attr(MPI.TAG_UB) - 1

        recvobj = sendobj
        mask = 1

        while mask < size:
            if (mask & rank) != 0:
                target = (rank & ~mask) % size
                self.send(recvobj, dest=target, tag=tag)
            else:
                target = rank | mask
                if target < size:
                    tmp = self.recv(None, source=target, tag=tag)
                    recvobj = op(recvobj, tmp)
            mask <<= 1

        if root != 0:
            if rank == 0:
                self.send(recvobj, dest=root, tag=tag)
            elif rank == root:
                recvobj = self.recv(None, source=0, tag=tag)

        if rank != root:
            recvobj = None

        return recvobj

    def allreduce(self, sendobj=None, recvobj=None, op=MPI.SUM):
        recvobj = self.reduce(sendobj, recvobj, op, 0)
        recvobj = self.bcast(recvobj, 0)
        return recvobj

    def scan(self, sendobj=None, recvobj=None, op=MPI.SUM):
        size = self.size
        rank = self.rank
        tag = MPI.COMM_WORLD.Get_attr(MPI.TAG_UB) - 1

        recvobj = sendobj
        partial = sendobj
        mask = 1

        while mask < size:
            target = rank ^ mask
            if target < size:
                tmp = self.sendrecv(
                    partial,
                    dest=target,
                    source=target,
                    sendtag=tag,
                    recvtag=tag,
                )
                if rank > target:
                    partial = op(tmp, partial)
                    recvobj = op(tmp, recvobj)
                else:
                    tmp = op(partial, tmp)
                    partial = tmp
            mask <<= 1

        return recvobj

    def exscan(self, sendobj=None, recvobj=None, op=MPI.SUM):
        size = self.size
        rank = self.rank
        tag = MPI.COMM_WORLD.Get_attr(MPI.TAG_UB) - 1

        recvobj = sendobj
        partial = sendobj
        mask = 1
        flag = False

        while mask < size:
            target = rank ^ mask
            if target < size:
                tmp = self.sendrecv(
                    partial,
                    dest=target,
                    source=target,
                    sendtag=tag,
                    recvtag=tag,
                )
                if rank > target:
                    partial = op(tmp, partial)
                    if rank != 0:
                        if not flag:
                            recvobj = tmp
                            flag = True
                        else:
                            recvobj = op(tmp, recvobj)
                else:
                    tmp = op(partial, tmp)
                    partial = tmp
            mask <<= 1

        if rank == 0:
            recvobj = None

        return recvobj



#!/usr/bin/env python
# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""mpi4py: Python bindings for MPI."""

# ruff: noqa: C408
# ruff: noqa: D103
# ruff: noqa: PTH100
# ruff: noqa: PTH118
# ruff: noqa: PTH120
# ruff: noqa: PTH207

import glob
import os
import sys

topdir = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.join(topdir, "conf"))

# --------------------------------------------------------------------
# Metadata
# --------------------------------------------------------------------

require_python = (3, 8)
maxknow_python = (3, 14)
py_limited_api = (3, 10)


def get_metadata():
    import metadata as md

    req_py = ">={}.{}".format(*require_python)
    assert req_py == md.requires_python  # noqa: S101
    author = md.authors[0]
    readme = md.get_readme()
    return {
        # distutils
        "name": md.get_name(),
        "version": md.get_version(),
        "description": md.description,
        "long_description": readme["text"],
        "classifiers": md.classifiers,
        "keywords": md.keywords,
        "license": md.license,
        "author": author["name"],
        "author_email": author["email"],
        # setuptools
        "project_urls": md.urls,
        "python_requires": md.requires_python,
        "long_description_content_type": readme["content-type"],
    }


def get_build_mpiabi():
    abi = os.environ.get("MPI4PY_BUILD_MPIABI", "").lower()
    return abi in {"true", "yes", "on", "y", "1"}


def get_build_pysabi():
    abi = os.environ.get("MPI4PY_BUILD_PYSABI", "").lower()
    if abi and sys.implementation.name == "cpython":
        if abi in {"false", "no", "off", "n", "0"}:
            return None
        if abi in {"true", "yes", "on", "y", "1"} | {"abi3"}:
            return py_limited_api
        if abi.startswith("cp"):
            abi = abi[2:]
        if "." in abi:
            x, y = abi.split(".")
        else:
            x, y = abi[0], abi[1:]
        return (int(x), int(y))
    return None


# --------------------------------------------------------------------
# Extension modules
# --------------------------------------------------------------------


def sources():
    # mpi4py.MPI
    MPI = dict(
        source="src/mpi4py/MPI.pyx",
        depends=[
            "src/mpi4py/*.pyx",
            "src/mpi4py/*.pxd",
            "src/mpi4py/MPI.src/*.pyx",
            "src/mpi4py/MPI.src/*.pxi",
        ],
    )
    #
    return [MPI]


def extensions():
    import mpidistutils

    # MPI extension module
    MPI = dict(
        name="mpi4py.MPI",
        sources=["src/mpi4py/MPI.c"],
        depends=(
            glob.glob("src/*.h")
            + glob.glob("src/lib-mpi/*.h")
            + glob.glob("src/lib-mpi/config/*.h")
            + glob.glob("src/lib-mpi/compat/*.h")
        ),
        include_dirs=["src"],
        define_macros=[],
        configure=mpidistutils.configure_mpi,
    )
    if get_build_mpiabi():
        MPI["define_macros"] += [
            ("PYMPIABI", 1),
        ]
    if sys.version_info[:2] > maxknow_python:
        MPI["define_macros"] += [
            ("CYTHON_LIMITED_API", 1),
        ]
    #
    return [MPI]


def executables():
    import mpidistutils

    # MPI-enabled Python interpreter
    pyexe = dict(
        name="python-mpi",
        optional=True,
        package="mpi4py",
        dest_dir="bin",
        sources=["src/python.c"],
        configure=mpidistutils.configure_pyexe,
    )
    #
    return [pyexe]


# --------------------------------------------------------------------
# Setup
# --------------------------------------------------------------------

package_info = dict(
    packages=[
        "mpi4py",
        "mpi4py.futures",
        "mpi4py.util",
    ],
    package_data={
        "mpi4py": [
            "*.pxd",
            "MPI*.h",
            "include/mpi4py/*.h",
            "include/mpi4py/*.i",
            "include/mpi4py/*.pxi",
            "py.typed",
            "*.pyi",
            "*/*.pyi",
        ],
    },
    package_dir={"": "src"},
)


def run_setup():
    """Call setuptools.setup(*args, **kwargs)."""
    try:
        import setuptools
    except ImportError as exc:
        setuptools = None
        if sys.version_info >= (3, 12):
            sys.exit(exc)
    from mpidistutils import (
        Executable as Exe,
        Extension as Ext,
        build_src,
        setup,
    )

    #
    build_src.sources = sources()
    #
    metadata = get_metadata()
    builder_args = dict(
        ext_modules=[Ext(**ext) for ext in extensions()],
        executables=[Exe(**exe) for exe in executables()],
    )
    if setuptools:
        builder_args["zip_safe"] = False
    else:
        metadata.pop("project_urls")
        metadata.pop("python_requires")
        metadata.pop("long_description_content_type")
    #
    sabi = get_build_pysabi()
    if sabi:
        api_ver = "0x{:02X}{:02X}0000".format(*sabi)
        defines = [("Py_LIMITED_API", api_ver)]
        for ext in builder_args["ext_modules"]:
            ext.define_macros.extend(defines)
            ext.py_limited_api = True
    if sabi and setuptools:
        api_tag = "cp{}{}".format(*sabi)
        options = {"bdist_wheel": {"py_limited_api": api_tag}}
        builder_args["options"] = options
    #
    setup_args = dict(
        i
        for d in (
            metadata,
            package_info,
            builder_args,
        )
        for i in d.items()
    )
    #
    setup(**setup_args)


def run_skbuild():
    """Call setuptools.setup(*args, **kwargs)."""
    from setuptools import setup

    #
    metadata = get_metadata()
    builder_args = dict(
        cmake_source_dir=".",
    )
    #
    options = {}
    cmake_args = []
    if get_build_mpiabi():
        cmake_args += ["-DMPIABI:BOOL=1"]
    sabi = get_build_pysabi()
    if sabi:
        api_tag = "cp{}{}".format(*sabi)
        options["bdist_wheel"] = {"py_limited_api": api_tag}
        cmake_args += ["-DPYSABI:STRING={}.{}".format(*sabi)]
    if options:
        builder_args["options"] = options
    if cmake_args:
        builder_args["cmake_args"] = cmake_args
    #
    setup_args = dict(
        i
        for d in (
            metadata,
            package_info,
            builder_args,
        )
        for i in d.items()
    )
    #
    setup(**setup_args)


# --------------------------------------------------------------------


def main():
    try:
        name = __import__("builder").get_build_backend_name()
    except RuntimeError as exc:
        sys.exit(exc)

    if name == "setuptools":
        run_setup()
    if name == "skbuild":
        run_skbuild()


if __name__ == "__main__":
    if sys.version_info < require_python:
        raise SystemExit(
            "error: requires Python version "
            + ".".join(map(str, require_python))
        )
    main()


# --------------------------------------------------------------------



import math

from mpi4py.futures import MPIPoolExecutor

PRIMES = [
    112272535095293,
    112582705942171,
    112272535095293,
    115280095190773,
    115797848077099,
    117450548693743,
    993960000099397,
]


def is_prime(n):
    if n % 2 == 0:
        return False
    sqrt_n = math.floor(math.sqrt(n))
    for i in range(3, sqrt_n + 1, 2):
        if n % i == 0:
            return False
    return True


def test_primes():
    with MPIPoolExecutor(4) as executor:
        for number, prime in zip(PRIMES, executor.map(is_prime, PRIMES)):
            print(f"{number:d} is prime: {prime}")


if __name__ == "__main__":
    test_primes()



import os
import platform

import arrayimpl
import mpiunittest as unittest

from mpi4py import MPI


def allclose(a, b, rtol=1.0e-5, atol=1.0e-8):
    try:
        iter(a)
    except TypeError:
        a = [a]
    try:
        iter(b)
    except TypeError:
        b = [b]
    for x, y in zip(a, b):
        if x == y:
            continue
        if abs(x - y) > (atol + rtol * abs(y)):
            return False
    return True


class BaseTestPack:
    #
    COMM = MPI.COMM_NULL

    skipdtype = []

    def testPackSize(self):
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                if typecode in self.skipdtype:
                    continue
                datatype = array.TypeMap[typecode]
                itemsize = datatype.Get_size()
                overhead = datatype.Pack_size(0, self.COMM)
                for count in range(10):
                    with self.subTest(count=count):
                        pack_size = datatype.Pack_size(count, self.COMM)
                        self.assertEqual(
                            pack_size - overhead, count * itemsize
                        )

    def testPackUnpack(self):
        for array, typecode1 in arrayimpl.loop():
            with arrayimpl.test(self):
                if typecode1 in self.skipdtype:
                    continue
                for typecode2 in array.TypeMap:
                    if typecode2 in self.skipdtype:
                        continue
                    datatype1 = array.TypeMap[typecode1]
                    datatype2 = array.TypeMap[typecode2]
                    for count in range(10):
                        with self.subTest(
                            typecode=(typecode1, typecode2), count=count
                        ):
                            # input and output arrays
                            iarray1 = array(range(count), typecode1).as_raw()
                            iarray2 = array(range(count), typecode2).as_raw()
                            oarray1 = array(count, typecode1, count).as_raw()
                            oarray2 = array(count, typecode2, count).as_raw()
                            # temp array for packing
                            size1 = datatype1.Pack_size(
                                len(iarray1), self.COMM
                            )
                            size2 = datatype2.Pack_size(
                                len(iarray2), self.COMM
                            )
                            tmpbuf = array(0, "b", size1 + size2 + 1).as_raw()
                            # pack input arrays
                            position = 0
                            position = datatype1.Pack(
                                iarray1, tmpbuf, position, self.COMM
                            )
                            position = datatype2.Pack(
                                iarray2, tmpbuf, position, self.COMM
                            )
                            # unpack output arrays
                            position = 0
                            position = datatype1.Unpack(
                                tmpbuf, position, oarray1, self.COMM
                            )
                            position = datatype2.Unpack(
                                tmpbuf, position, oarray2, self.COMM
                            )
                            # test
                            self.assertTrue(allclose(iarray1, oarray1))
                            self.assertTrue(allclose(iarray2, oarray2))


EXT32 = "external32"


class BaseTestPackExternal:
    #
    skipdtype = []

    def testPackSize(self):
        for array, typecode in arrayimpl.loop():
            with arrayimpl.test(self):
                if typecode in self.skipdtype:
                    continue
                datatype = array.TypeMap[typecode]
                overhead = datatype.Pack_external_size(EXT32, 0)
                for count in range(10):
                    with self.subTest(count=count):
                        pack_size = datatype.Pack_external_size(EXT32, count)
                        real_size = pack_size - overhead
                        self.assertGreaterEqual(real_size, 0)

    @unittest.skipMPI("openmpi(<5.0.0)", platform.machine() == "sparc64")
    def testPackUnpackExternal(self):
        for array, typecode1 in arrayimpl.loop():
            with arrayimpl.test(self):
                if unittest.is_mpi_gpu("mpich", array):
                    continue
                if unittest.is_mpi_gpu("openmpi", array):
                    continue
                if unittest.is_mpi_gpu("mvapich", array):
                    continue
                if typecode1 in self.skipdtype:
                    continue
                for typecode2 in array.TypeMap:
                    if typecode2 in self.skipdtype:
                        continue
                    datatype1 = array.TypeMap[typecode1]
                    datatype2 = array.TypeMap[typecode2]
                    for count in range(1, 10):
                        with self.subTest(
                            count=count, typecode=(typecode1, typecode2)
                        ):
                            # input and output arrays
                            val = 127 if typecode1 == "b" else 255
                            iarray1 = array(val, typecode1, count).as_raw()
                            iarray2 = array(range(count), typecode2).as_raw()
                            oarray1 = array(-1, typecode1, count).as_raw()
                            oarray2 = array(-1, typecode2, count).as_raw()
                            # temp array for packing
                            size1 = datatype1.Pack_external_size(
                                EXT32, len(iarray1)
                            )
                            size2 = datatype2.Pack_external_size(
                                EXT32, len(iarray2)
                            )
                            tmpbuf = array(0, "b", size1 + size2 + 1).as_raw()
                            # pack input arrays
                            position = 0
                            position = datatype1.Pack_external(
                                EXT32, iarray1, tmpbuf, position
                            )
                            position = datatype2.Pack_external(
                                EXT32, iarray2, tmpbuf, position
                            )
                            # unpack output arrays
                            position = 0
                            position = datatype1.Unpack_external(
                                EXT32, tmpbuf, position, oarray1
                            )
                            position = datatype2.Unpack_external(
                                EXT32, tmpbuf, position, oarray2
                            )
                            # test result
                            self.assertTrue(allclose(iarray1, oarray1))
                            self.assertTrue(allclose(iarray2, oarray2))


class TestPackSelf(BaseTestPack, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


class TestPackWorld(BaseTestPack, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


@unittest.skipMPI("openmpi(<3.0.0)")
class TestPackExternal(BaseTestPackExternal, unittest.TestCase):
    #
    pass


name, version = MPI.get_vendor()
if name == "MPICH":
    if version < (4, 0, 0):
        BaseTestPackExternal.skipdtype += "ldgLFDG"
    if platform.architecture(None)[0] == "32bit":
        BaseTestPackExternal.skipdtype += "gG"
elif name == "Open MPI":
    if version < (5, 0, 0):
        BaseTestPackExternal.skipdtype += "gG"
    if (platform.system(), platform.machine()) == ("Darwin", "arm64"):
        BaseTestPackExternal.skipdtype += "G"
elif name == "Intel MPI":
    BaseTestPackExternal.skipdtype += "lgLG"
    BaseTestPackExternal.skipdtype += "D"
    if os.name == "nt":
        BaseTestPackExternal.skipdtype += "q"
elif name == "Microsoft MPI":
    BaseTestPackExternal.skipdtype += "gFDG"
elif name == "MVAPICH":
    BaseTestPackExternal.skipdtype += "lfdgLFDG"
elif name == "MPICH2":
    BaseTestPackExternal.skipdtype += "ldgLFDG"
else:
    try:
        MPI.BYTE.Pack_external_size(EXT32, 0)
    except NotImplementedError:
        unittest.disable(BaseTestPackExternal, "mpi-ext32")


if __name__ == "__main__":
    unittest.main()



import arrayimpl
import mpiunittest as unittest

from mpi4py import MPI


def skip_op(typecode, op):
    if typecode == "?":
        return True
    if typecode in "FDG":
        if op in (MPI.MAX, MPI.MIN):
            return True
    return False


def maxvalue(a):
    try:
        typecode = a.typecode
    except AttributeError:
        typecode = a.dtype.char
    if typecode == ("f"):
        return 1e30
    elif typecode == ("d"):
        return 1e300
    else:
        return 2 ** (a.itemsize * 7) - 1


@unittest.skipMPI("openmpi(<1.6.0)")
@unittest.skipMPI("msmpi", MPI.COMM_WORLD.Get_size() >= 3)
@unittest.skipMPI("MPICH1")
@unittest.skipIf(MPI.ROOT == MPI.PROC_NULL, "mpi-root")
@unittest.skipIf(MPI.COMM_WORLD.Get_size() < 2, "mpi-world-size<2")
class BaseTestCCOBufInter:
    #
    BASECOMM = MPI.COMM_NULL
    INTRACOMM = MPI.COMM_NULL
    INTERCOMM = MPI.COMM_NULL

    def setUp(self):
        size = self.BASECOMM.Get_size()
        rank = self.BASECOMM.Get_rank()
        if rank < size // 2:
            self.COLOR = 0
            self.LOCAL_LEADER = 0
            self.REMOTE_LEADER = size // 2
        else:
            self.COLOR = 1
            self.LOCAL_LEADER = 0
            self.REMOTE_LEADER = 0
        self.INTRACOMM = self.BASECOMM.Split(self.COLOR, key=0)
        Create_intercomm = MPI.Intracomm.Create_intercomm
        self.INTERCOMM = Create_intercomm(
            self.INTRACOMM,
            self.LOCAL_LEADER,
            self.BASECOMM,
            self.REMOTE_LEADER,
        )

    def tearDown(self):
        self.INTRACOMM.Free()
        self.INTERCOMM.Free()

    def testBarrier(self):
        self.INTERCOMM.Barrier()

    def testBcast(self):
        comm = self.INTERCOMM
        rank = comm.Get_rank()
        size = comm.Get_size()
        rsize = comm.Get_remote_size()
        for array, typecode in arrayimpl.loop():
            for color in (0, 1):
                if self.COLOR == color:
                    for root in range(size):
                        if root == rank:
                            buf = array(root, typecode, root + color)
                            comm.Bcast(buf.as_mpi(), root=MPI.ROOT)
                        else:
                            comm.Bcast(None, root=MPI.PROC_NULL)
                else:
                    for root in range(rsize):
                        buf = array(-1, typecode, root + color)
                        comm.Bcast(buf.as_mpi(), root=root)
                        check = arrayimpl.scalar(root)
                        for value in buf:
                            self.assertEqual(value, check)

    def testGather(self):
        comm = self.INTERCOMM
        rank = comm.Get_rank()
        size = comm.Get_size()
        rsize = comm.Get_remote_size()
        for array, typecode in arrayimpl.loop():
            for color in (0, 1):
                if self.COLOR == color:
                    for root in range(size):
                        if root == rank:
                            rbuf = array(-1, typecode, (rsize, root + color))
                            comm.Gather(None, rbuf.as_mpi(), root=MPI.ROOT)
                            check = arrayimpl.scalar(root)
                            for value in rbuf.flat:
                                self.assertEqual(value, check)
                        else:
                            comm.Gather(None, None, root=MPI.PROC_NULL)
                else:
                    for root in range(rsize):
                        sbuf = array(root, typecode, root + color)
                        comm.Gather(sbuf.as_mpi(), None, root=root)

    def testScatter(self):
        comm = self.INTERCOMM
        rank = comm.Get_rank()
        size = comm.Get_size()
        rsize = comm.Get_remote_size()
        for array, typecode in arrayimpl.loop():
            for color in (0, 1):
                if self.COLOR == color:
                    for root in range(size):
                        if root == rank:
                            sbuf = array(root, typecode, (rsize, root + color))
                            comm.Scatter(sbuf.as_mpi(), None, root=MPI.ROOT)
                        else:
                            comm.Scatter(None, None, root=MPI.PROC_NULL)
                else:
                    for root in range(rsize):
                        rbuf = array(root, typecode, root + color)
                        comm.Scatter(None, rbuf.as_mpi(), root=root)
                        check = arrayimpl.scalar(root)
                        for value in rbuf:
                            self.assertEqual(value, check)

    def testAllgather(self):
        comm = self.INTERCOMM
        comm.Get_rank()
        size = comm.Get_size()
        rsize = comm.Get_remote_size()
        for array, typecode in arrayimpl.loop():
            for color in (0, 1):
                if self.COLOR == color:
                    for n in range(size):
                        sbuf = array(n, typecode, color)
                        rbuf = array(-1, typecode, (rsize, n + color))
                        comm.Allgather(sbuf.as_mpi(), rbuf.as_mpi())
                        check = arrayimpl.scalar(n)
                        for value in rbuf.flat:
                            self.assertEqual(value, check)
                else:
                    for n in range(rsize):
                        sbuf = array(n, typecode, n + color)
                        rbuf = array(-1, typecode, (rsize, color))
                        comm.Allgather(sbuf.as_mpi(), rbuf.as_mpi())
                        check = arrayimpl.scalar(n)
                        for value in rbuf.flat:
                            self.assertEqual(value, check)

    def testAlltoall(self):
        comm = self.INTERCOMM
        comm.Get_rank()
        size = comm.Get_size()
        rsize = comm.Get_remote_size()
        for array, typecode in arrayimpl.loop():
            for color in (0, 1):
                if self.COLOR == color:
                    for n in range(size):
                        sbuf = array(n, typecode, (rsize, (n + 1) * color))
                        rbuf = array(-1, typecode, (rsize, n + 3 * color))
                        comm.Alltoall(sbuf.as_mpi(), rbuf.as_mpi())
                        check = arrayimpl.scalar(n)
                        for value in rbuf.flat:
                            self.assertEqual(value, check)
                else:
                    for n in range(rsize):
                        sbuf = array(n, typecode, (rsize, n + 3 * color))
                        rbuf = array(-1, typecode, (rsize, (n + 1) * color))
                        comm.Alltoall(sbuf.as_mpi(), rbuf.as_mpi())
                        check = arrayimpl.scalar(n)
                        for value in rbuf.flat:
                            self.assertEqual(value, check)

    @unittest.skipMPI("mvapich", MPI.COMM_WORLD.Get_size() > 2)
    def testReduce(self):
        comm = self.INTERCOMM
        rank = comm.Get_rank()
        lsize = comm.Get_size()
        rsize = comm.Get_remote_size()
        for array, typecode in arrayimpl.loop():
            for op in (MPI.SUM, MPI.PROD, MPI.MAX, MPI.MIN):
                if skip_op(typecode, op):
                    continue
                for color in (0, 1):
                    if self.COLOR == color:
                        for root in range(lsize):
                            if root == rank:
                                rbuf = array(-1, typecode, rsize)
                                comm.Reduce(
                                    None,
                                    rbuf.as_mpi(),
                                    op=op,
                                    root=MPI.ROOT,
                                )
                                max_val = maxvalue(rbuf)
                                for i, value in enumerate(rbuf):
                                    if op == MPI.SUM:
                                        if (i * rsize) < max_val:
                                            self.assertAlmostEqual(
                                                value, i * rsize
                                            )
                                    elif op == MPI.PROD:
                                        if (i**rsize) < max_val:
                                            self.assertAlmostEqual(
                                                value, i**rsize
                                            )
                                    elif op == MPI.MAX:
                                        self.assertEqual(value, i)
                                    elif op == MPI.MIN:
                                        self.assertEqual(value, i)
                            else:
                                comm.Reduce(
                                    None,
                                    None,
                                    op=op,
                                    root=MPI.PROC_NULL,
                                )
                    else:
                        for root in range(rsize):
                            sbuf = array(range(lsize), typecode)
                            comm.Reduce(
                                sbuf.as_mpi(),
                                None,
                                op=op,
                                root=root,
                            )

    def testAllreduce(self):
        comm = self.INTERCOMM
        comm.Get_rank()
        comm.Get_size()
        rsize = comm.Get_remote_size()
        for array, typecode in arrayimpl.loop():
            for op in (MPI.SUM, MPI.PROD, MPI.MAX, MPI.MIN):
                if skip_op(typecode, op):
                    continue
                sbuf = array(range(5), typecode)
                rbuf = array([-1] * 5, typecode)
                comm.Allreduce(sbuf.as_mpi(), rbuf.as_mpi(), op)
                max_val = maxvalue(rbuf)
                for i, value in enumerate(rbuf):
                    if op == MPI.SUM:
                        if (i * rsize) < max_val:
                            self.assertAlmostEqual(value, i * rsize)
                    elif op == MPI.PROD:
                        if (i**rsize) < max_val:
                            self.assertAlmostEqual(value, i**rsize)
                    elif op == MPI.MAX:
                        self.assertEqual(value, i)
                    elif op == MPI.MIN:
                        self.assertEqual(value, i)


class TestCCOBufInter(BaseTestCCOBufInter, unittest.TestCase):
    #
    BASECOMM = MPI.COMM_WORLD


class TestCCOBufInterDup(TestCCOBufInter):
    #
    def setUp(self):
        self.BASECOMM = self.BASECOMM.Dup()
        super().setUp()

    def tearDown(self):
        self.BASECOMM.Free()
        super().tearDown()


if __name__ == "__main__":
    unittest.main()



import ctypes
import pathlib

_libdir = pathlib.Path(__file__).resolve().parent

MPI_Fint = ctypes.c_int


class MPI_Comm(ctypes.Structure):
    _fields_ = [("mpi_val", MPI_Fint)]


lib = ctypes.CDLL(_libdir / "libhelloworld.so")
lib.sayhello.restype = None
lib.sayhello.argtypes = [ctypes.POINTER(MPI_Comm)]


def sayhello(comm):
    comm_f = MPI_Comm(comm.py2f())
    lib.sayhello(comm_f)



# Very, very naive RE-based way for collecting declarations inside
# 'cdef extern from *' Cython blocks in in source files, and next
# generate compatibility headers for partially implemented MPIs.

# ruff: noqa: UP031

import pathlib
import re
import sys
import warnings
from textwrap import dedent, indent


def anyof(*args):
    return r"(?:{})".format("|".join(args))


def join(*args):
    tokens = []
    for tok in args:
        if isinstance(tok, (list, tuple)):
            tok = "({})".format(r"\s*".join(tok))
        tokens.append(tok)
    return r"\s*".join(tokens)


def r_(*args):
    return re.compile(join(r"^", *args, r"$"))


lparen = r"\("
rparen = r"\)"
colon = r"\:"
asterisk = r"\*"
opt = r"?"

enum = join("enum", colon)
typedef = "ctypedef"
pointer = asterisk
struct = join(typedef, "struct")

integral_type_names = [
    "Aint",
    "Offset",
    "Count",
    "Fint",
]

struct_type_names = [
    "Status",
    "F08_status",
]

handle_type_names = [
    "Datatype",
    "Request",
    "Message",
    "Op",
    "Info",
    "Group",
    "Errhandler",
    "Session",
    "Comm",
    "Win",
    "File",
]

basic_type = r"(?:void|int|char\s*\*{1,3})"
integral_type = r"MPI_(?:{})".format("|".join(integral_type_names))
struct_type = r"MPI_(?:{})".format("|".join(struct_type_names))
opaque_type = r"MPI_(?:{})".format("|".join(handle_type_names))

upper_name = r"MPI_[A-Z0-9_]+"
camel_name = r"MPI_[A-Z][a-z0-9_]+"
usrfun_name = camel_name + r"_(?:function|function_c|fn)"

arg_list = r".*"
ret_type = r"void|int|double|MPI_Aint"

canyint = anyof(r"int", r"long(?:\s+long)?")
canyptr = join(r"\w+", pointer + "?")

annotation = r"\#\:\="
fallback_value = r"\(?[A-Za-z0-9_\+\-\(\)\*]+\)?"
fallback = rf"(?:{join(annotation, [fallback_value])})?"

cint_type = r"int"
cmpi_type = opaque_type.replace("Datatype", "Type")
h2i_name = cmpi_type + "_toint"
i2h_name = cmpi_type + "_fromint"

fint_type = r"MPI_Fint"
fmpi_type = opaque_type.replace("Datatype", "Type")
c2f_name = fmpi_type + "_c2f"
f2c_name = fmpi_type + "_f2c"


class Re:
    INTEGRAL_TYPE = r_(typedef, [canyint], [integral_type], fallback)
    STRUCT_TYPE = r_(struct, [struct_type], colon + opt, fallback)
    OPAQUE_TYPE = r_(typedef, canyptr, [opaque_type])
    FUNCTION_TYPE = r_(
        typedef, [ret_type], [camel_name], lparen, [arg_list], rparen, fallback
    )

    ENUM_VALUE = r_(enum, [upper_name], fallback)
    HANDLE_VALUE = r_([opaque_type], [upper_name], fallback)
    BASIC_PTRVAL = r_([basic_type, pointer], [upper_name], fallback)
    INTEGRAL_PTRVAL = r_([integral_type, pointer], [upper_name], fallback)
    STRUCT_PTRVAL = r_([struct_type, pointer], [upper_name], fallback)
    FUNCTION_PTRVAL = r_([usrfun_name, pointer], [upper_name], fallback)

    FUNCTION_PROTO = r_(
        [ret_type], [camel_name], lparen, [arg_list], rparen, fallback
    )
    FUNCTION_H2I = r_(
        [cint_type], [h2i_name], lparen, [opaque_type], rparen, fallback
    )
    FUNCTION_I2H = r_(
        [opaque_type], [i2h_name], lparen, [cint_type], rparen, fallback
    )
    FUNCTION_C2F = r_(
        [fint_type], [c2f_name], lparen, [opaque_type], rparen, fallback
    )
    FUNCTION_F2C = r_(
        [opaque_type], [f2c_name], lparen, [fint_type], rparen, fallback
    )

    IGNORE = r_(
        anyof(
            join(r"cdef.*"),
            join(struct, r"_mpi_\w+_t"),
            join(r"int", r"MPI_(?:SOURCE|TAG|ERROR)"),
            join(r"#.*"),
            join(r""),
        )
    )


class Node:
    REGEX = None

    @classmethod
    def match(self, line):
        m = self.REGEX.match(line)
        return m.groups() if m else None

    HEADER = None
    CONFIG = None
    MISSING = None

    MISSING_HEAD = """\
    #ifndef PyMPI_HAVE_%(name)s
    #undef  %(cname)s
    """
    MISSING_TAIL = """
    #endif

    """

    def init(self, name, **kwargs):
        self.name = name
        self.__dict__.update(kwargs)

    def header(self):
        line = dedent(self.HEADER) % vars(self)
        line = line.replace("\n", "")
        line = line.replace("  ", " ")
        return line + "\n"

    def config(self):
        return dedent(self.CONFIG) % vars(self)

    def missing(self, guard=True):
        if guard:
            head = dedent(self.MISSING_HEAD)
            tail = dedent(self.MISSING_TAIL)
        else:
            head = "#undef  %(cname)s\n"
            tail = "\n\n"
        body = dedent(self.MISSING)
        return (head + body + tail) % vars(self)


class NodeType(Node):
    CONFIG = """\
    %(ctype)s v; %(ctype)s* p; (void)v; (void)p;"""

    def __init__(self, ctype):
        self.init(
            name=ctype,
            cname=ctype,
            ctype=ctype,
        )


class NodeStructType(NodeType):
    HEADER = """\
    typedef struct {%(cfields)s ...; } %(ctype)s;"""
    MISSING = """\
    typedef struct PyMPI_%(ctype)s {
    %(cfields)s
    } PyMPI_%(ctype)s;
    #define %(ctype)s PyMPI_%(ctype)s"""

    def __init__(self, ctype, cfields):
        super().__init__(ctype)
        self.cfields = "\n".join([
            f"  {ctype} {cname};" for ctype, cname in cfields
        ])


class NodeFuncType(NodeType):
    HEADER = """\
    typedef %(crett)s (%(cname)s)(%(cargs)s);"""
    MISSING = """\
    typedef %(crett)s (MPIAPI PyMPI_%(cname)s)(%(cargs)s);
    #define %(cname)s PyMPI_%(cname)s"""

    def __init__(self, crett, cname, cargs, calias=None):
        self.init(
            name=cname,
            cname=cname,
            ctype=cname + "*",
        )
        self.crett = crett
        self.cargs = cargs or "void"
        if calias is not None:
            self.MISSING = "#define %(cname)s %(calias)s"
            self.calias = calias


class NodeValue(Node):
    HEADER = """\
    const %(ctype)s %(cname)s;"""
    CONFIG = """\
    %(ctype)s v; v = %(cname)s; (void)v;"""
    MISSING = "#define %(cname)s (%(calias)s)"

    def __init__(self, ctype, cname, calias):
        self.init(name=cname, cname=cname, ctype=ctype, calias=calias)
        if ctype.endswith("*"):
            ctype = ctype + " const"
            self.HEADER = ctype + " %(cname)s;"


class NodePtrVal(NodeValue):
    MISSING = "#define %(cname)s ((%(ctype)s)%(calias)s)"


def ctypefix(ct):
    ct = ct.replace("[][3]", " (*)[3]")
    ct = ct.replace("[]", "*")
    return ct


FALLBACK = {
    "MPI_Aint_add": "(a0,a1) ((MPI_Aint)((char*)(a0)+(a1)))",
    "MPI_Aint_diff": "(a0,a1) ((MPI_Aint)((char*)(a0)-(char*)(a1)))",
    "MPI_Wtime": "() 0.0",
    "MPI_Wtick": "() 0.0",
}


class NodeFuncProto(Node):
    HEADER = """\
    %(crett)s %(cname)s(%(cargs)s);"""
    CONFIG = """\
    %(crett)s v; v = %(cname)s(%(cargscall)s); (void)v;"""
    MISSING = " ".join([
        "#define %(cname)s(%(cargsnamed)s)",
        'PyMPI_UNAVAILABLE("%(name)s"%(comma)s%(cargsnamed)s)',
    ])

    def __init__(self, crett, cname, cargs, calias=None):
        self.init(name=cname, cname=cname)
        self.crett = crett
        self.cargs = cargs or "void"
        if cargs == "void":
            cargs = ""
        if cargs:
            cargs = [c.strip() for c in cargs.split(",")]
        else:
            cargs = []
        self.cargstype = cargs[:]
        cargs = [a for a in cargs if a != "..."]
        nargs = len(cargs)
        self.comma = "," if nargs else ""
        cargscall = [f"({ctypefix(a)})0" for a in cargs]
        self.cargscall = ",".join(cargscall)
        cargsnamed = ["a%d" % (a + 1) for a in range(nargs)]
        self.cargsnamed = ",".join(cargsnamed)
        if calias is not None:
            self.calias = calias
            self.MISSING = "#define %(cname)s %(calias)s"
        elif cname in FALLBACK:
            self.MISSING = "#define %(cname)s" + FALLBACK[cname]


class IntegralType(NodeType):
    REGEX = Re.INTEGRAL_TYPE
    HEADER = """\
    typedef %(cbase)s... %(ctype)s;"""
    MISSING = """\
    typedef %(ctdef)s PyMPI_%(ctype)s;
    #define %(ctype)s PyMPI_%(ctype)s"""

    def __init__(self, cbase, ctype, calias=None):
        super().__init__(ctype)
        self.cbase = cbase
        if calias is not None:
            self.ctdef = calias
        else:
            self.ctdef = cbase


class StructType(NodeStructType):
    REGEX = Re.STRUCT_TYPE

    def __init__(self, ctype, calias=None):
        cfields = []
        if ctype == "MPI_Status":
            cnames = ["MPI_SOURCE", "MPI_TAG", "MPI_ERROR"]
            cfields = list(zip(["int"] * 3, cnames))
        super().__init__(ctype, cfields)
        if calias is not None:
            self.MISSING = "#define %(cname)s %(calias)s"
            self.calias = calias


class OpaqueType(NodeType):
    REGEX = Re.OPAQUE_TYPE
    HEADER = """\
    typedef struct{...;} %(ctype)s;"""
    MISSING = """\
    typedef void *PyMPI_%(ctype)s;
    #define %(ctype)s PyMPI_%(ctype)s"""


class FunctionType(NodeFuncType):
    REGEX = Re.FUNCTION_TYPE


class EnumValue(NodeValue):
    REGEX = Re.ENUM_VALUE

    def __init__(self, cname, calias):
        self.init(name=cname, cname=cname, ctype="int", calias=calias)


class HandleValue(NodeValue):
    REGEX = Re.HANDLE_VALUE
    MISSING = "#define %(cname)s ((%(ctype)s)%(calias)s)"


class BasicPtrVal(NodePtrVal):
    REGEX = Re.BASIC_PTRVAL


class IntegralPtrVal(NodePtrVal):
    REGEX = Re.INTEGRAL_PTRVAL


class StructPtrVal(NodePtrVal):
    REGEX = Re.STRUCT_PTRVAL


class FunctionPtrVal(NodePtrVal):
    REGEX = Re.FUNCTION_PTRVAL


class FunctionProto(NodeFuncProto):
    REGEX = Re.FUNCTION_PROTO


class FunctionH2I(NodeFuncProto):
    REGEX = Re.FUNCTION_H2I
    MISSING = """\
    #ifdef  PyMPI_HAVE_%(fallback)s
    #define %(cname)s(%(cargsnamed)s) ((%(crett)s)%(fallback)s(%(cargsnamed)s))
    #else
    #define %(cname)s(%(cargsnamed)s) ((void)%(cargsnamed)s,%(cretv)s)
    #endif"""

    def __init__(self, *a, **k):
        NodeFuncProto.__init__(self, *a, **k)
        self.fallback = self.name.replace("_toint", "_c2f")
        self.cretv = f"({self.crett})-1"


class FunctionI2H(NodeFuncProto):
    REGEX = Re.FUNCTION_I2H
    MISSING = """\
    #ifdef  PyMPI_HAVE_%(fallback)s
    #define %(cname)s(%(cargsnamed)s) (%(fallback)s((%(cargs)s)%(cargsnamed)s))
    #else
    #define %(cname)s(%(cargsnamed)s) ((void)%(cargsnamed)s,%(cretv)s)
    #endif"""

    def __init__(self, *a, **k):
        NodeFuncProto.__init__(self, *a, **k)
        self.fallback = self.name.replace("_fromint", "_f2c")
        self.cretv = f"({self.crett})0"


class FunctionC2F(NodeFuncProto):
    REGEX = Re.FUNCTION_C2F
    MISSING = FunctionH2I.MISSING

    def __init__(self, *a, **k):
        NodeFuncProto.__init__(self, *a, **k)
        self.fallback = self.name.replace("_c2f", "_toint")
        self.cretv = f"({self.crett})-1"


class FunctionF2C(NodeFuncProto):
    REGEX = Re.FUNCTION_F2C
    MISSING = FunctionI2H.MISSING

    def __init__(self, *a, **k):
        NodeFuncProto.__init__(self, *a, **k)
        self.fallback = self.name.replace("_f2c", "_fromint")
        self.cretv = f"({self.crett})0"


class Generator:
    PROLOG = f"""\
    /* Generated with `python conf/{pathlib.Path(__file__).name}` */
    """

    NODE_TYPES = [
        IntegralType,
        StructType,
        OpaqueType,
        FunctionType,
        HandleValue,
        EnumValue,
        BasicPtrVal,
        IntegralPtrVal,
        StructPtrVal,
        FunctionPtrVal,
        FunctionH2I,
        FunctionI2H,
        FunctionC2F,
        FunctionF2C,
        FunctionProto,
    ]

    def __init__(self):
        self.nodes = []
        self.nodemap = {}
        self.stdapi = {}

    def parse_stdapi(self, filename):
        filename = pathlib.Path(filename)
        header = re.compile(r'^\s*#include\s+"mpi-(\d\d)\.h"\s*$')
        define = re.compile(
            r"^\s*#define\s*PyMPI_HAVE_(MPI_[A-Za-z0-9_]+)\s+1\s*$"
        )
        with filename.open(encoding="utf-8") as f:
            for line in f:
                m = header.match(line)
                if m is None:
                    continue
                numversion = int(m.groups()[0])
                version = (numversion // 10, numversion % 10)
                self.stdapi[version] = symlist = []
                mpiheader = filename.parent / f"mpi-{numversion}.h"
                with mpiheader.open(encoding="utf-8") as h:
                    for line in h:
                        m = define.match(line)
                        if m is None:
                            continue
                        symbol = m.groups()[0]
                        index = self.nodemap.get(symbol)
                        if index is None:
                            continue
                        node = self.nodes[index]
                        node.version = version
                        symlist.append(symbol)
        for a, la in self.stdapi.items():
            for b, lb in self.stdapi.items():
                if a != b:
                    common = set(la).intersection(set(lb))
                    if common:
                        raise AssertionError(f"{a} & {b} -> {common}")

    def parse_file(self, filename):
        filename = pathlib.Path(filename)
        with filename.open(encoding="utf-8") as f:
            self.parse_lines(f)

    def parse_lines(self, lines):
        for line in lines:
            self.parse_line(line)

    def parse_line(self, line):
        if "# Deprecated" in line:
            self.deprecated = True
            return
        elif Re.IGNORE.match(line):
            self.deprecated = False
            return
        nodemap = self.nodemap
        nodelist = self.nodes
        deprecated = getattr(self, "deprecated", False)
        for nodetype in self.NODE_TYPES:
            args = nodetype.match(line)
            if args:
                node = nodetype(*args)
                nodemap[node.name] = len(nodelist)
                nodelist.append(node)
                node.deprecated = deprecated
                break
        if not args:
            warnings.warn(f"unmatched line:\n{line}", stacklevel=1)

    def __iter__(self):
        return iter(self.nodes)

    def __getitem__(self, name):
        return self.nodes[self.nodemap[name]]

    def dump_header_h(self, fileobj):
        if isinstance(fileobj, (str, pathlib.Path)):
            fileobj = pathlib.Path(fileobj)
            with fileobj.open("w", encoding="utf-8") as f:
                self.dump_header_h(f)
            return
        for node in self:
            fileobj.write(node.header())

    CONFIG_HEAD = """\
    #ifndef PyMPI_PYMPICONF_H
    #define PyMPI_PYMPICONF_H

    """
    CONFIG_MACRO = "PyMPI_HAVE_%s"
    CONFIG_TAIL = """\

    #endif /* PyMPI_PYMPICONF_H */
    """

    def dump_config_h(self, fileobj, suite):
        if isinstance(fileobj, (str, pathlib.Path)):
            fileobj = pathlib.Path(fileobj)
            with fileobj.open("w", encoding="utf-8") as f:
                self.dump_config_h(f, suite)
            return
        prlg = dedent(self.PROLOG)
        head = dedent(self.CONFIG_HEAD)
        macro = dedent(self.CONFIG_MACRO)
        tail = dedent(self.CONFIG_TAIL)
        fileobj.write(prlg)
        fileobj.write(head)
        if suite is None:
            for node in self:
                line = "#undef %s\n" % (macro % node.name)
                fileobj.write(line)
        else:
            for name, result in suite:
                if result:
                    line = "#define %s 1\n" % (macro % name)
                else:
                    line = "#undef  %s\n" % (macro % name)
                fileobj.write(line)
        fileobj.write(tail)

    MISSING_HEAD = """
    """
    MISSING_TAIL = """\
    /* */
    """

    def dump_missing_h(self, fileobj, suite=None):
        if isinstance(fileobj, (str, pathlib.Path)):
            fileobj = pathlib.Path(fileobj)
            with fileobj.open("w", encoding="utf-8") as f:
                self.dump_missing_h(f, suite)
            return
        prlg = dedent(self.PROLOG)
        head = dedent(self.MISSING_HEAD)
        tail = dedent(self.MISSING_TAIL)
        #
        fileobj.write(prlg)
        fileobj.write(head)
        if suite is None:
            for name in (
                "MPI_Status_c2f",
                "MPI_Status_f2c",
                "MPI_Type_create_f90_integer",
                "MPI_Type_create_f90_real",
                "MPI_Type_create_f90_complex",
            ):
                fileobj.write(
                    dedent(f"""\
                #ifdef PyMPI_MISSING_{name}
                #undef PyMPI_HAVE_{name}
                #endif
                \n""")
                )
            for node in self:
                fileobj.write(node.missing())
        else:
            for name, result in suite:
                node = self[name]
                if not result:
                    fileobj.write(node.missing())
        fileobj.write(tail)

    LARGECNT_HEAD = """
    #define PyMPIAllocArray(dsttype, dst, len)                       \\
      do {                                                           \\
          size_t _m = (size_t) (len) * sizeof(dsttype);              \\
          (dst) = (dsttype *) PyMPI_MALLOC(_m ? _m : 1);             \\
      } while (0)                                                 /**/

    #define PyMPIFreeArray(dst)                                      \\
      do {                                                           \\
        if ((dst) != NULL) PyMPI_FREE(dst);                          \\
        (dst) = NULL; (void) (dst);                                  \\
      } while (0)                                                 /**/

    #define PyMPICastError(ERRORCODE)                                \\
      do {                                                           \\
        ierr = (ERRORCODE);                                          \\
        (void) MPI_Comm_call_errhandler(MPI_COMM_SELF, ierr);        \\
        goto fn_exit;                                                \\
      } while (0)                                                 /**/

    #define PyMPICastValue(dsttype, dst, srctype, src)               \\
      do {                                                           \\
        (dst) = (dsttype) (src);                                     \\
        if ((srctype) (dst) != (src))                                \\
          PyMPICastError(MPI_ERR_ARG);                               \\
      } while (0)                                                 /**/

    #define PyMPICastArray(dsttype, dst, srctype, src, len)          \\
      do {                                                           \\
        (dst) = NULL;                                                \\
        if ((src) != NULL) {                                         \\
          MPI_Aint _n = (MPI_Aint) (len), _i;                        \\
          PyMPIAllocArray(dsttype, dst, len);                        \\
          if ((dst) == NULL)                                         \\
            PyMPICastError(MPI_ERR_OTHER);                           \\
          for (_i = 0; _i < _n; _i++) {                              \\
            (dst)[_i] = (dsttype) (src)[_i];                         \\
            if ((srctype) (dst)[_i] != (src)[_i]) {                  \\
              PyMPIFreeArray(dst);                                   \\
              PyMPICastError(MPI_ERR_ARG);                           \\
            }                                                        \\
          }                                                          \\
        }                                                            \\
      } while (0)                                                 /**/

    #define PyMPIMoveArray(dsttype, dst, srctype, src, len)          \\
      do {                                                           \\
        if ((src) != NULL && (dst) != NULL) {                        \\
          size_t _n = (size_t) (len);                                \\
          unsigned char *_buf = (unsigned char *) (src);             \\
          (void) PyMPI_MEMCPY(_buf, (dst), _n * sizeof(dsttype));    \\
          PyMPI_FREE(dst); (dst) = (dsttype *) _buf;                 \\
        }                                                            \\
      } while (0)                                                 /**/

    #define PyMPICommSize(comm, n)                                   \\
      do {                                                           \\
        int _inter = 0;                                              \\
        ierr = MPI_Comm_test_inter(comm, &_inter);                   \\
        if (_inter)                                                  \\
          ierr = MPI_Comm_remote_size((comm), &(n));                 \\
        else                                                         \\
          ierr = MPI_Comm_size((comm), &(n));                        \\
        if (ierr != MPI_SUCCESS) goto fn_exit;                       \\
      } while (0)                                                 /**/

    #define PyMPICommLocalGroupSize(comm, n)                         \\
      do {                                                           \\
        ierr = MPI_Comm_size((comm), &(n));                          \\
        if (ierr != MPI_SUCCESS) goto fn_exit;                       \\
      } while (0)                                                 /**/

    #define PyMPICommNeighborCount(comm, ns, nr)                     \\
      do {                                                           \\
        int _topo = MPI_UNDEFINED;                                   \\
        int _i, _n; (ns) = (nr) = 0;                                 \\
        ierr = MPI_Topo_test((comm), &_topo);                        \\
        if (ierr != MPI_SUCCESS) goto fn_exit;                       \\
        if (_topo == MPI_UNDEFINED) {                                \\
          ierr = MPI_Comm_size((comm), &_n);                         \\
          (ns) = (nr) = _n;                                          \\
        } else if (_topo == MPI_CART) {                              \\
          ierr = MPI_Cartdim_get((comm), &_n);                       \\
          (ns) = (nr) = 2 * _n;                                      \\
        } else if (_topo == MPI_GRAPH) {                             \\
          ierr = MPI_Comm_rank((comm), &_i);                         \\
          ierr = MPI_Graph_neighbors_count(                          \\
                   (comm), _i, &_n);                                 \\
          (ns) = (nr) = _n;                                          \\
        } else if (_topo == MPI_DIST_GRAPH) {                        \\
          ierr = MPI_Dist_graph_neighbors_count(                     \\
                   (comm), &(nr), &(ns), &_i);                       \\
        }                                                            \\
        if (ierr != MPI_SUCCESS) goto fn_exit;                       \\
      } while (0)                                                 /**/

    """

    LARGECNT_BEGIN = """\
    #if !defined(PyMPI_HAVE_%(name)s_c) || PyMPI_LEGACY_ABI
    #undef %(name)s_c
    """

    LARGECNT_DECLARE = """\
    static int Py%(name)s_c(%(argsdecl)s)
    {
      PyMPI_WEAK_CALL(%(name)s_c, %(argsorig)s);
      {
    """

    LARGECNT_COLLECTIVE = """\
    PyMPICommSize(a%(commid)d, n);
    """

    LARGECNT_LOCALGROUP = """\
    PyMPICommLocalGroupSize(a%(commid)d, n);
    """

    LARGECNT_NEIGHBOR = """\
      PyMPICommNeighborCount(a%(commid)d, ns, nr);
    """

    LARGECNT_CALL = """\
    ierr = %(callname)s(%(argscall)s);
    if (ierr != MPI_SUCCESS) goto fn_exit;
    """

    LARGECNT_RETURN = """\
      return ierr;
      }
    }
    """

    LARGECNT_END = """\
    #undef  %(name)s_c
    #define %(name)s_c Py%(name)s_c
    #endif
    """

    LARGECNT_TAIL = """\
    /* */
    """

    LARGECNT_RE = re.compile(r"^mpi_({})_c$".format("|".join([
        # r"type_contiguous", # implemented
        r"type_(vector|indexed|create|size|get_(true_)?extent).*",
        r"(get|status_set)_elements|get_count",
        r"(pack|unpack)(_external)?(_size)?",
        r"(i?(b|s|r|p)?send(_init)?(recv(_replace)?)?)",
        r"(i?m?p?recv(_init)?)",
        r"(buffer_(at|de)tach)|(comm|session)_(at|de)tach_buffer",
        r"(i?(bcast|gather(v)?|scatter(v?))(_init)?)",
        r"(i?(all(gather(v)?|toall(v|w)?))(_init)?)",
        r"(i?((all)?reduce(_local|(_scatter(_block)?)?)|(ex)?scan)(_init)?)",
        r"(i?neighbor_all(gather(v)?|toall(v|w)?)(_init)?)",
        r"(win_(create|allocate(_shared)?|shared_query))",
        r"(r?(put|get|(get_)?accumulate))",
        r"file_(((i)?(read|write).*)|get_type_extent)",
    ])))  # fmt: skip

    def dump_largecnt_h(self, fileobj):
        if isinstance(fileobj, (str, pathlib.Path)):
            fileobj = pathlib.Path(fileobj)
            with fileobj.open("w", encoding="utf-8") as f:
                self.dump_largecnt_h(f)
            return

        def largecount_functions():
            for node in self:
                name = node.name
                if self.LARGECNT_RE.match(name.lower()):
                    yield name[:-2]

        def declare(t, v, init=None):
            t = t.strip()
            if t.endswith("[]"):
                t = t[:-2].strip()
                code = f"{t} *{v}"
            elif t.endswith("*"):
                t = t[:-1].strip()
                code = f"{t} *{v}"
            else:
                code = f"{t} {v}"
            if init is not None:
                code += f" = {init}"
            return code

        def generate(name):
            subname = name[4:].lower()
            is_neighbor = "neighbor" in subname
            is_nonblocking = name.startswith("MPI_I")
            is_datatype = subname.startswith("type_")
            is_packunpack = any(map(subname.startswith, ("pack", "unpack")))
            is_packunpack_size = is_packunpack and subname.endswith("_size")

            node1 = self[name + "_c"]
            node2 = self[name]
            try:
                node2 = self[name + "_x"]
            except KeyError:
                pass
            callname = node2.name

            cargstype1 = node1.cargstype
            cargstype2 = node2.cargstype
            argstype = list(zip(cargstype1, cargstype2))
            convert_array = False
            for t1, t2 in argstype:
                if t1 != t2:
                    if t1 == "MPI_Count[]" and t2 == "int[]":
                        convert_array = True
                        break
                    if t1 == "MPI_Aint[]" and t2 == "int[]":
                        convert_array = True
                        break
            commid = None
            if convert_array:
                for i, (t1, _) in enumerate(argstype, start=1):
                    if t1 == "MPI_Comm":
                        commid = i
                        break

            dtypeidx = 0
            argslist = []
            argsorig = []
            argsinit = []
            argstemp = []
            argsconv = []
            argscall = []
            argsoutp = []
            argsfree = []

            CASTVALUE = "PyMPICastValue(%s, b%d, %s, a%d)"
            CASTARRAY = "PyMPICastArray(%s, b%d, %s, a%d, %s)"
            MOVEARRAY = "PyMPIMoveArray(%s, b%d, %s, a%d, %s)"
            FREEARRAY = "PyMPIFreeArray(b%d)"

            argsinit += ["int ierr"]
            if commid is not None:
                if is_neighbor:
                    argsinit += ["int ns, nr"]
                else:
                    argsinit += ["int n"]
            for i, (t1, t2) in enumerate(argstype, start=1):
                argslist += [declare(t1, "a%d" % i)]
                argsorig += ["a%d" % i]
                if t1.startswith("MPI_Datatype"):
                    dtypeidx += 1
                if t1 == t2:
                    argscall += ["a%d" % i]
                else:
                    if t1.endswith("[]"):
                        t1, t2, n = t1[:-2], t2[:-2], "n"
                        argstemp += [declare(t2, "*b%d" % i, "NULL")]
                        if is_datatype:
                            is_darray = name.endswith("_darray")
                            n = "a1" if not is_darray else "a3"
                        if is_neighbor:
                            n = ("ns", "nr")[dtypeidx]
                        subs = (t2, i, t1, i, n)
                        argsconv += [CASTARRAY % subs]
                        if is_nonblocking:
                            argsconv += [MOVEARRAY % subs]
                        else:
                            argsfree += [FREEARRAY % i]
                        argscall += ["b%d" % i]
                    elif t1.endswith("*"):
                        t1, t2 = t1[:-1], t2[:-1]
                        pinit = "a%d ? &b%d : NULL" % (i, i)
                        argstemp += [declare(t2, "b%d" % i, 0)]
                        argstemp += [declare(t2 + "*", "p%d" % i, pinit)]
                        argscall += ["p%d" % i]
                        argsoutp += ["if (a%d) *a%d = b%d" % (i, i, i)]
                        if is_packunpack and not is_packunpack_size:
                            subs = (t2, i, t1, f"a{i} ? *a{i} : 0")
                            castvalue = CASTVALUE.replace("a%d", "%s")
                            argsconv += [castvalue % subs]
                    else:
                        subs = (t2, i, t1, i)
                        argstemp += [declare(t2, "b%d" % i)]
                        argsconv += [CASTVALUE % subs]
                        argscall += ["b%d" % i]

            tab = "  "
            subs = {
                "name": name,
                "callname": callname,
                "argsdecl": (",\n" + " " * (len(name) + 20)).join(argslist),
                "argsorig": ", ".join(argsorig),
                "argscall": ", ".join(argscall),
                "commid": commid,
            }
            begin = self.LARGECNT_BEGIN % subs
            fdecl = self.LARGECNT_DECLARE % subs
            if commid is None:
                setup = ""
            elif is_neighbor:
                setup = self.LARGECNT_NEIGHBOR % subs
            elif "reduce_scatter" in name.lower():
                setup = self.LARGECNT_LOCALGROUP % subs
            else:
                setup = self.LARGECNT_COLLECTIVE % subs
            call = self.LARGECNT_CALL % subs
            ret = self.LARGECNT_RETURN % subs
            end = self.LARGECNT_END % subs

            yield dedent(begin)
            yield dedent(fdecl)
            if argsinit:
                yield indent("{};\n".format("; ".join(argsinit)), tab)
            if argstemp:
                yield indent("{};\n".format("; ".join(argstemp)), tab)
            yield indent(dedent(setup), tab)
            for line in argsconv:
                yield indent(f"{line};\n", tab)
            yield indent(dedent(call), tab)
            for line in argsoutp:
                yield indent(f"{line};\n", tab)
            yield indent("fn_exit:\n", tab[:-1])
            for line in argsfree:
                yield indent(f"{line};\n", tab)
            yield dedent(ret)
            yield dedent(end)
            yield "\n"

        prlg = dedent(self.PROLOG)
        head = dedent(self.LARGECNT_HEAD)
        tail = dedent(self.LARGECNT_TAIL)
        self.largecnt = 0
        fileobj.write(prlg)
        fileobj.write(head)
        for name in largecount_functions():
            self.largecnt += 1
            for code in generate(name):
                fileobj.write(code)
        fileobj.write(tail)

    MPIABI_HEAD = """
    #define _pympi_CALL(fn, ...) \\
    PyMPI_WEAK_CALL(fn, __VA_ARGS__); \\
    return _pympi__##fn(__VA_ARGS__)
    """
    MPIABI_TAIL = """
    #undef _pympi_CALL
    /* */
    """

    def dump_mpiabi_h(self, fileobj, std=True):
        if isinstance(fileobj, (str, pathlib.Path)):
            fileobj = pathlib.Path(fileobj)
            with fileobj.open("w", encoding="utf-8") as f:
                self.dump_mpiabi_h(f, std)
            return

        def funcargs(node):
            argsdecl, argscall = [], []
            for i, t in enumerate(node.cargstype):
                if t == "...":
                    argsdecl.append(t)
                else:
                    try:
                        p = t.index("[")
                    except ValueError:
                        p = len(t)
                    t, a = t[:p], t[p:]
                    argsdecl.append(f"{t} a{i}{a}")
                    argscall.append(f"a{i}")
            if node.name in ("MPI_Status_c2f", "MPI_Status_f2c"):
                if not argsdecl[0].startswith("const "):
                    argsdecl[0] = "const " + argsdecl[0]
            argsdecl = ",".join(argsdecl) if argsdecl else "void"
            argscall = ",".join(argscall) if argscall else ""
            return (argsdecl, argscall)

        def abi_function(node, fallback=None, guard=True):
            name = node.name

            rtype = node.crett
            rvalue = f'PyMPI_UNAVAILABLE("{name}")'
            if rtype != "int":
                rvalue = f"(({rtype})0)"
            rvalue = getattr(node, "cretv", rvalue)
            argsdecl, argscall = funcargs(node)

            pympi0 = f"{name}"
            pympi1 = f"_pympi_{name}"
            fsign0 = f"{rtype} {pympi0}({argsdecl})"
            fsign1 = f"{rtype} {pympi1}({argsdecl})"
            fbody1 = f"{{ _pympi_CALL({name},{argscall}); }}"

            undef = dedent(f"""
            #undef {name}
            """)
            if guard:
                declare = dedent(f"""\
                #ifndef PyMPI_HAVE_{name}
                PyMPI_EXTERN {fsign0};
                #endif
                """)
            else:
                declare = dedent(f"""\
                PyMPI_EXTERN {fsign0};
                """)

            if fallback is None:
                pympiname = f"_pympi__{name}"
                fallback = dedent(f"""\
                #define {pympiname}(...) {rvalue}
                """)
            override = dedent(f"""\
            PyMPI_LOCAL {fsign1} {fbody1}
            #define {name} {pympi1}
            """)
            return [undef, declare, fallback, override]

        def abi_handle_openmpi(node):
            name = node.name.lower()
            oname = f"ompi_{name}"
            if node.ctype == "MPI_Datatype":
                if oname in (
                    "ompi_mpi_long_long",  # ompi_mpi_long_long_int
                    "ompi_mpi_c_complex",  # ompi_mpi_c_float_complex
                    "ompi_mpi_bfloat16_t",  # unavailable
                ):
                    return ""
                for old, new in (
                    ("complex", "cplex"),
                    ("double_complex", "dblcplex"),
                    ("double_precision", "dblprec"),
                    ("long_double_int", "longdbl_int"),
                    ("cxx_float_complex", "cxx_cplex"),
                    ("cxx_double_complex", "cxx_dblcplex"),
                    ("cxx_long_double_complex", "cxx_ldblcplex"),
                    ("float16_t", "short_float"),
                ):
                    if oname == f"ompi_mpi_{old}":
                        oname = f"ompi_mpi_{new}"
            if node.ctype == "MPI_Op":
                if not oname.endswith("_null"):
                    oname = oname.replace("_mpi_", "_mpi_op_")
            for htype in ("message", "request"):
                if name.startswith(f"mpi_{htype}_"):
                    oname = f"o{name}"
            if node.ctype == "MPI_Session":
                oname = oname.replace("session", "instance")
            return dedent(f"""\
            PyMPI_WEAK_LOAD({oname})
            """)

        mpi_version_min = (5, 0) if std else (3, 0)
        mpi_version_max = (5, 0)

        ftnconv = []
        intconv = []
        handles = []
        fstatus = []
        fortran = []
        aintops = []
        functions = []
        for node in self:
            if node.deprecated:
                continue
            if isinstance(node, (FunctionH2I, FunctionI2H)):
                intconv.append(node)
                continue
            if isinstance(node, (FunctionC2F, FunctionF2C)):
                ftnconv.append(node)
                continue
            if isinstance(node, HandleValue):
                handles.append(node)
                continue
            if isinstance(node, FunctionProto):
                if node.name in ("MPI_Status_c2f", "MPI_Status_f2c"):
                    fstatus.append(node)
                    continue
                if node.name.startswith("MPI_Type_create_f90_"):
                    fortran.append(node)
                    continue
                if node.name in ("MPI_Aint_add", "MPI_Aint_diff"):
                    aintops.append(node)
                    continue
                functions.append(node)
                continue

        prlg = dedent(self.PROLOG)
        head = dedent(self.MPIABI_HEAD)
        tail = dedent(self.MPIABI_TAIL)
        fileobj.write(prlg)
        fileobj.write(head)

        if not std:
            fileobj.write(
                dedent("""
            #ifdef OPEN_MPI
            #include <mpi-ext.h>
            """)
            )
            for node in handles:
                code = abi_handle_openmpi(node)
                fileobj.write(code)
            fileobj.write(
                dedent("""\
            #endif /* OPEN_MPI */
            """)
            )

        if not std:
            fileobj.write(
                dedent("""
            #ifdef MPICH
            """)
            )
            for node in fortran:
                for code in abi_function(node):
                    fileobj.write(code)
            fileobj.write(
                dedent("""
            #endif /* MPICH */
            """)
            )

        if not std:
            argdcl = {"c": "MPI_Status *c", "f": "MPI_Fint *f"}
            argval = {"c": "*c", "f": "*(MPI_Status *)(char *)f"}
            for node in fstatus:
                name = node.name
                pympiname = f"_pympi__{name}"
                a, _, b = name[-3:].partition("2")
                args = f"(const {argdcl[a]}, {argdcl[b]})"
                copy = f"({argval[b]} = {argval[a]})"
                ok, err = "MPI_SUCCESS", "MPI_ERR_ARG"
                body = f"{{ return (c && f) ? {copy}, {ok} : {err}; }}"
                impl = dedent(f"""\
                static int {pympiname}{args} {body}
                """)
                for code in abi_function(node, impl, guard=False):
                    fileobj.write(code)
            for node in ftnconv:
                name = node.name
                pympiname = f"_pympi__{name}"
                rtype = node.crett
                rval_o = node.cretv
                rval_m = f"({rtype})arg"
                if name.startswith("MPI_File_"):
                    rval_m = node.cretv
                impl = dedent(f"""\
                #ifdef MPICH
                #define {pympiname}(arg) ({rval_m})
                #else
                #define {pympiname}(arg) ({rval_o})
                #endif
                """)
                for code in abi_function(node, impl, guard=False):
                    fileobj.write(code)
        if not std:
            for node in intconv:
                name = node.name
                pympiname = f"_pympi__{name}"
                if node.crett == "int":
                    rtype, atype = "(int)", ""
                else:
                    rtype, atype = "", "(int)"
                body = f"{rtype}{node.fallback}({atype}arg)"
                impl = dedent(f"""\
                #define {pympiname}(arg) {body}
                """)
                for code in abi_function(node, impl):
                    fileobj.write(code)

        if not std:
            for node in aintops:
                name = node.name
                pympiname = f"_pympi__{name}"
                rtype = node.crett
                argsdecl, argscall = funcargs(node)
                signature = f"{rtype} {pympiname}({argsdecl})"
                expr1 = f"{name}({argscall})"
                expr2 = FALLBACK[name].partition(" ")[2]
                impl = dedent(f"""
                #ifdef {name}
                static {signature} {{ return {expr1}; }}
                #else
                static {signature} {{ return {expr2}; }}
                #endif
                """)
                code = abi_function(node, impl, guard=False)
                undef, declare, fallback, override = code
                code = [fallback, undef.lstrip(), declare, override]
                while code:
                    fileobj.write(code.pop(0))

        outdir = pathlib.Path(fileobj.name).parent
        undef_re = re.compile(r"^\s*#undef\s+(MPI_.*)\s*$")
        implemented = set()
        for header in outdir.glob("mpiapi??.h"):
            with header.open(encoding="utf-8") as f:
                for line in f:
                    match = undef_re.match(line)
                    if match:
                        implemented.add(match.group(1))
        for node in functions:
            if node.name.endswith("_c"):
                implemented.add(node.name)
        implemented.discard("MPI_Op_create_c")

        for node in functions:
            if node.version <= mpi_version_min:
                continue
            if node.version > mpi_version_max:
                continue
            code = abi_function(node)
            fileobj.write(code.pop(0))
            fileobj.write(code.pop(0))
            if node.name not in implemented:
                while code:
                    fileobj.write(code.pop(0))

        fileobj.write(tail)


# -----------------------------------------

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="MPI API generator")
    parser.color = True  # Python 3.14+
    parser.add_argument("-q", "--quiet", action="store_true")
    parser.add_argument("-l", "--list", action="store_true")
    args = parser.parse_args()

    if args.list:
        args.quiet = True

    def log(message):
        if not args.quiet:
            print(message)

    topdir = pathlib.Path()

    generator = Generator()
    libmpi_pxd = topdir / "src" / "mpi4py" / "libmpi.pxd"
    log(f"parsing file {libmpi_pxd}")
    generator.parse_file(libmpi_pxd)
    log("processed %d definitions" % len(generator.nodes))
    mpiapi_h = topdir / "src" / "lib-mpi" / "config" / "mpiapi.h"
    log(f"parsing file {mpiapi_h}")
    generator.parse_stdapi(mpiapi_h)
    log("processed %d definitions" % sum(map(len, generator.stdapi.values())))

    if args.list:
        for node in generator:
            print(node.name)
        sys.exit(0)

    # config_h  = topdir / "src" / "lib-mpi" / "pympiconf.h"
    # log("writing file %s" % config_h)
    # generator.dump_config_h(config_h, None)

    missing_h = topdir / "src" / "lib-mpi" / "missing.h"
    log(f"writing file {missing_h}")
    generator.dump_missing_h(missing_h)

    mpiabi0_h = topdir / "src" / "lib-mpi" / "mpiabi0.h"
    log(f"writing file {mpiabi0_h}")
    generator.dump_mpiabi_h(mpiabi0_h, std=False)

    mpiabi1_h = topdir / "src" / "lib-mpi" / "mpiabi1.h"
    log(f"writing file {mpiabi1_h}")
    generator.dump_mpiabi_h(mpiabi1_h, std=True)

    largecnt_h = topdir / "src" / "lib-mpi" / "largecnt.h"
    log(f"writing file {largecnt_h}")
    generator.dump_largecnt_h(largecnt_h)
    log("generated %d large count fallbacks" % generator.largecnt)

    # libmpi_h = topdir / ".", "libmpi.h"
    # log("writing file %s" % libmpi_h)
    # generator.dump_header_h(libmpi_h)

# -----------------------------------------



# cython: language_level=3str

# ---------

# Python-level module import
# (file: mpi4py/MPI.*.so)

from mpi4py import MPI

# Python-level objects and code

size  = MPI.COMM_WORLD.Get_size()
rank  = MPI.COMM_WORLD.Get_rank()
pname = MPI.Get_processor_name()

hwmess = "Hello, World! I am process %d of %d on %s."
print (hwmess % (rank, size, pname))

# ---------

# Cython-level cimport
# this make available mpi4py's Python extension types
# (file:  mpi4py/MPI.pxd)

from mpi4py cimport MPI
from mpi4py.MPI cimport Intracomm as IntracommType

from mpi4py.libmpi cimport MPI_Comm_size
from mpi4py.libmpi cimport MPI_Comm_rank
from mpi4py.libmpi cimport MPI_MAX_PROCESSOR_NAME
from mpi4py.libmpi cimport MPI_Get_processor_name

# C-level cdef, typed, Python objects

cdef MPI.Comm WORLD = MPI.COMM_WORLD
cdef IntracommType SELF = MPI.COMM_SELF

cdef int ierr0 = 0

cdef int size0 = 0
ierr0 = MPI_Comm_size(WORLD.ob_mpi, &size0)

cdef int rank0 = 0
ierr0 = MPI_Comm_rank(WORLD.ob_mpi, &rank0)

cdef int rlen0 = 0
cdef char pname0[MPI_MAX_PROCESSOR_NAME]
ierr0 = MPI_Get_processor_name(pname0, &rlen0)
pname0[rlen0] = 0  # just in case ;-)

hwmess = "Hello, World! I am process %d of %d on %s."
print (hwmess % (rank0, size0, pname0.decode()))

# ---------

# Cython-level cimport with PXD file
# this make available the native MPI C API
# with namespace-protection (stuff accessed as mpi.XXX)
# (file: mpi4py/libmpi.pxd)

from mpi4py cimport libmpi as mpi

cdef mpi.MPI_Comm world1 = WORLD.ob_mpi

cdef int ierr1 = 0

cdef int size1 = 0
ierr1 = mpi.MPI_Comm_size(mpi.MPI_COMM_WORLD, &size1)

cdef int rank1 = 0
ierr1 = mpi.MPI_Comm_rank(mpi.MPI_COMM_WORLD, &rank1)

cdef int rlen1 = 0
cdef char pname1[mpi.MPI_MAX_PROCESSOR_NAME]
ierr1 = mpi.MPI_Get_processor_name(pname1, &rlen1)
pname1[rlen1] = 0  # just in case ;-)

hwmess = "Hello, World! I am process %d of %d on %s."
print (hwmess % (rank1, size1, pname1.decode()))

# ---------



"""
Compare the speed of primes sequentially vs. using futures.
"""

import math
import sys
import time

try:
    from concurrent.futures import ThreadPoolExecutor
except ImportError:
    ThreadPoolExecutor = None
try:
    from concurrent.futures import ProcessPoolExecutor
except ImportError:
    ProcessPoolExecutor = None

from mpi4py.futures import MPIPoolExecutor

PRIMES = [
    112272535095293,
    112582705942171,
    112272535095293,
    115280095190773,
    115797848077099,
    117450548693743,
    993960000099397,
]


def is_prime(n):
    if n % 2 == 0:
        return False
    sqrt_n = math.floor(math.sqrt(n))
    for i in range(3, sqrt_n + 1, 2):
        if n % i == 0:
            return False
    return True


def sequential():
    return list(map(is_prime, PRIMES))


def with_thread_pool_executor():
    if not ThreadPoolExecutor:
        return None
    with ThreadPoolExecutor(4) as executor:
        return list(executor.map(is_prime, PRIMES))


def with_process_pool_executor():
    if not ProcessPoolExecutor:
        return None
    with ProcessPoolExecutor(4) as executor:
        return list(executor.map(is_prime, PRIMES))


def with_mpi_pool_executor():
    with MPIPoolExecutor(4) as executor:
        return list(executor.map(is_prime, PRIMES))


def main():
    for name, fn in [
        ("sequential", sequential),
        ("threads", with_thread_pool_executor),
        ("processes", with_process_pool_executor),
        ("mpi4py", with_mpi_pool_executor),
    ]:
        sys.stdout.write(f"{name.ljust(11)}: ")
        sys.stdout.flush()
        start = time.time()
        result = fn()
        if result is None:
            sys.stdout.write(" not available\n")
        elif result != [True] * len(PRIMES):
            sys.stdout.write(" failed\n")
        else:
            elapsed = time.time() - start
            sys.stdout.write(f"{elapsed:5.2f} seconds\n")
        sys.stdout.flush()


if __name__ == "__main__":
    main()



# Exception Class
# ---------------

class Exception(RuntimeError):
    """
    Exception class.
    """

    def __init__(self, int ierr: int = SUCCESS, /) -> None:
        """Initialize self."""
        if ierr < MPI_SUCCESS: ierr = MPI_ERR_UNKNOWN
        RuntimeError.__init__(self, ierr)

    def __int__(self) -> int:
        """Return int(self)."""
        return self.args[0]

    def __bool__(self) -> bool:
        """Return bool(self)."""
        return <int>self != MPI_SUCCESS

    def __hash__(self) -> int:
        """Return hash(self)."""
        return hash(int(self))

    def __eq__(self, object other: object, /) -> bool:
        """Return self==value."""
        return int(self) == other

    def __ne__(self, object other: object, /) -> bool:
        """Return self!=other."""
        return int(self) != other

    def __lt__(self, object other: int, /) -> bool:
        """Return self<other."""
        return int(self) < other

    def __le__(self, object other: int, /) -> bool:
        """Return self<=other."""
        return int(self) <= other

    def __gt__(self, object other: int, /) -> bool:
        """Return self>other."""
        return int(self) > other

    def __ge__(self, object other: int, /) -> bool:
        """Return self>=other."""
        return int(self) >= other

    def __repr__(self) -> str:
        """Return repr(self)."""
        return f"MPI.Exception({int(self)})"

    def __str__(self) -> str:
        """Return str(self)."""
        if mpi_version < 4 and not mpi_active():
            return f"error code: {int(self)}"  # ~> legacy
        return self.Get_error_string()

    def Get_error_code(self) -> int:
        """
        Error code.
        """
        cdef int errorcode = self
        return errorcode

    def Get_error_class(self) -> int:
        """
        Error class.
        """
        cdef int errorcode = self
        cdef int errorclass = MPI_SUCCESS
        CHKERR( MPI_Error_class(errorcode, &errorclass) )
        return errorclass

    def Get_error_string(self) -> str:
        """
        Error string.
        """
        cdef int errorcode = self
        cdef char string[MPI_MAX_ERROR_STRING+1]
        cdef int resultlen = 0
        CHKERR( MPI_Error_string(errorcode, string, &resultlen) )
        return tompistr(string, resultlen)

    @property
    def error_code(self) -> int:
        """Error code."""
        return self.Get_error_code()

    @property
    def error_class(self) -> int:
        """Error class."""
        return self.Get_error_class()

    @property
    def error_string(self) -> str:
        """Error string."""
        return self.Get_error_string()


MPIException = Exception
